"""Contains callbacks.

Callbacks are lightweight event handlers, used mainly for logging.

To make a method of a class trigger callbacks, one needs to
decorate the class (or its base class) with ``@introduce_callbacks()`` and then
also decorate the method with ``@apply_callbacks``.

Callbacks can be registered by simply supplying them in the respective keyword argument:
::

    @regelum.main(callbacks=[...], ...)
    def my_app(config):
        ...

"""

import logging
from abc import ABC, abstractmethod

import mlflow

import regelum
import pandas as pd

import time
import datetime
import dill
import os
import git
import shutil


import matplotlib.pyplot as plt

import pkg_resources

from pathlib import Path

import sys
import filelock
import regelum.__internal.base

import numpy as np
from typing import Dict, Any

from regelum.event import Event
import torch


def is_in_debug_mode():
    return sys.gettrace() is not None


def disable_in_jupyter(method):
    def wrapper(self: Callback, *args, **kwargs):
        if self.is_jupyter:
            return lambda *args, **kwargs: None
        else:
            return method(self, *args, **kwargs)

    return wrapper


def detach(Attachee):
    """Create a duplicate of the provided regelum type with all callbacks detached from it."""

    if not hasattr(Attachee, "_attached"):
        return Attachee

    class Detachee(Attachee):
        _attached = []
        _real_name = Attachee.__name__

    return Detachee


def trigger(CallbackClass):
    """Decorate a callback class in such a way that its event handling is inherited by derived classes.

    Args:
        CallbackClass (type): class to be decorated

    Returns:
        CallbackClass: altered class that passes down its handlers to derived classes
            (regardless of whether handling methods are overriden)
    """

    class PassdownCallback(CallbackClass):
        def __call_passdown(self, obj, method, output):
            if (
                True
            ):  # self.ready(t):   # Currently, cooldowns don't work for PassdownCallbacks
                try:
                    if CallbackClass.is_target_event(self, obj, method, output, []):
                        CallbackClass.on_function_call(self, obj, method, output)
                        return True
                    return False
                    # self.on_trigger(PassdownCallback)
                    # self.trigger_cooldown(t)
                except regelum.RegelumExitException as e:
                    raise e
                except Exception as e:
                    self.log(
                        f"Callback {self.__class__.__name__} failed, when executing routines passed down from {CallbackClass.__name__}."
                    )
                    self.exception(e)
                    return False

        def on_function_call(self, *args, **kwargs):
            pass

    PassdownCallback.__name__ = CallbackClass.__name__
    return PassdownCallback


def is_interactive():
    import __main__ as main
    import sys

    return (
        (not hasattr(main, "__file__"))
        and (not is_in_debug_mode())
        and (not ("pytest" in sys.modules))
    )


class Callback(regelum.__internal.base.RegelumBase, ABC):
    """Base class for callbacks.

    Callback objects are used to perform in response to some method being called.
    """

    cooldown = None
    is_jupyter = is_interactive()

    def __init__(self, log_level="info", attachee=None):
        """Initialize a callback object.

        :param logger: A logging object that will be used to log messages.
        :type logger: logging.Logger
        :param log_level: The level at which messages should be logged.
        :type log_level: str
        """
        super().__init__()
        self.attachee = attachee
        self.log = self._metadata["logger"].__getattribute__(log_level)
        # TODO: FIX THIS. Setting the level is needed due to the fact that mlflow sql backend reinstantiates logger
        # Moreover, rubbish mlflow backend logs are generated. They are not needed for a common user
        self._metadata["logger"].setLevel(logging.INFO)
        self.exception = self._metadata["logger"].exception
        self.__last_trigger = 0.0

    @classmethod
    def register(cls, *args, launch=False, **kwargs):
        existing_callbacks = [
            type(callback) for callback in cls._metadata["main"].callbacks
        ]
        if cls not in existing_callbacks:
            callback_instance = cls(*args, **kwargs)
            if launch:
                callback_instance.on_launch()
            cls._metadata["main"].callbacks = [callback_instance] + cls._metadata[
                "main"
            ].callbacks

    @abstractmethod
    def is_target_event(self, obj, method, output, triggers):
        pass

    def is_done_collecting(self):
        return True

    def on_episode_done(
        self,
        scenario,
        episode_number,
        episodes_total,
        iteration_number,
        iterations_total,
    ):
        pass

    def on_trigger(self, caller):
        pass

    def on_iteration_done(
        self,
        scenario,
        episode_number,
        episodes_total,
        iteration_number,
        iterations_total,
    ):
        pass

    def ready(self, t):
        if not self.cooldown:
            return True
        if t - self.__last_trigger > self.cooldown:
            return True
        else:
            return False

    def trigger_cooldown(self, t):
        self.__last_trigger = t

    def on_launch(self):
        pass

    @abstractmethod
    def on_function_call(self, obj, method, output):
        pass

    def __call__(self, obj, method, output):
        t = time.time()
        triggers = []
        for base in self.__class__.__bases__:
            if hasattr(base, f"_PassdownCallback__call_passdown"):
                if base._PassdownCallback__call_passdown(self, obj, method, output):
                    triggers.append(base)
        if self.ready(t):
            try:
                if self.is_target_event(obj, method, output, triggers):
                    self.on_function_call(obj, method, output)
                    done = True
                    for base in self.__class__.__bases__:
                        if hasattr(base, f"_PassdownCallback__call_passdown"):
                            done = done and base.is_done_collecting(self)
                    done = done and self.is_done_collecting()
                    if done and triggers:
                        self.on_trigger(self.__class__)
                    self.trigger_cooldown(t)
            except regelum.RegelumExitException as e:
                raise e
            except Exception as e:
                self.log(f"Callback {self.__class__.__name__} failed.")
                self.exception(e)

    @classmethod
    def attach(cls, other):
        if hasattr(other, "_attached"):
            attached = other._attached + [cls]
        else:
            attached = [cls]

        class Attachee(other):
            _real_name = other.__name__
            _attached = attached

        return Attachee

    """
    @classmethod
    def detach(cls, other):
        if hasattr(other, "_attached"):
            attached = other._attached.copy()
            if cls not in attached:
                raise ValueError(f"Attempted to detach {cls.__name__}, but it was not attached.")
            attached.pop(attached.index(cls))
        else:
            raise ValueError(f"Attempted to detach {cls.__name__}, but it was not attached.")
        class Detachee(other):
            _real_name = other.__name__
            _attached = attached
        Detachee.__name__ = other.__name__
        return Detachee
    """

    def on_termination(self, res):
        pass


class OnEpisodeDoneCallback(Callback):
    """Callback responsible for logging and recording relevant data when an episode ends."""

    def __init__(self, *args, **kwargs):
        """Initialize an OnEpisodeDoneCallback instance."""
        super().__init__(*args, **kwargs)
        self.episode_counter = 0
        self.iteration_counter = 1

    def is_target_event(self, obj, method, output, triggers):
        return isinstance(obj, regelum.scenario.Scenario) and (
            method == "reload_scenario"
        )

    def on_function_call(self, obj, method, output):
        self.episode_counter += 1
        for callback in self._metadata["main"].callbacks:
            callback.on_episode_done(
                obj,
                self.episode_counter,
                obj.N_episodes,
                self.iteration_counter,
                obj.N_iterations,
            )

        if self.episode_counter == obj.N_episodes:
            self.episode_counter = 0
            self.iteration_counter += 1


class OnIterationDoneCallback(Callback):
    """Callback responsible for logging and recording relevant data when an iteration ends."""

    def __init__(self, *args, **kwargs):
        """Initialize an instance of OnIterationDoneCallback."""
        super().__init__(*args, **kwargs)
        self.iteration_counter = 0

    def is_target_event(self, obj, method, output, triggers):
        return isinstance(obj, regelum.scenario.Scenario) and (
            method == "reset_iteration"
        )

    def on_function_call(self, obj, method, output):
        self.iteration_counter += 1
        if obj.N_iterations > 1 and self.iteration_counter <= obj.N_iterations:
            for callback in self._metadata["main"].callbacks:
                callback.on_iteration_done(
                    obj,
                    obj.N_episodes,
                    obj.N_episodes,
                    self.iteration_counter,
                    obj.N_iterations,
                )


class ConfigDiagramCallback(Callback):
    """Callback responsible for constructing SUMMARY.html and relevant data, including the config diagram."""

    def on_function_call(self, *args, **kwargs):
        pass

    def is_target_event(self, obj, method, output, triggers):
        return False

    @staticmethod
    def __monitor_git():
        try:
            forbidden = False
            with filelock.FileLock(regelum.main.metadata["common_dir"] + "/diff.lock"):
                repo = git.Repo(
                    path=regelum.main.metadata["initial_working_directory"],
                    search_parent_directories=True,
                )
                commit_hash = repo.head.object.hexsha
                if repo.is_dirty(untracked_files=True):
                    commit_hash += (
                        ' <font color="red">(uncommitted/unstaged changes)</font>'
                    )
                    forbidden = (
                        "disallow_uncommitted" in regelum.main.config
                        and regelum.main.config.disallow_uncommitted
                        and not is_in_debug_mode()
                    )
                    untracked = repo.untracked_files
                    if not os.path.exists(
                        regelum.main.metadata["common_dir"] + "/changes.diff"
                    ):
                        repo.git.add(all=True)
                        with open(
                            regelum.main.metadata["common_dir"] + "/changes.diff", "w"
                        ) as f:
                            diff = repo.git.diff(repo.head.commit.tree)
                            if untracked:
                                repo.index.remove(untracked, cached=True)
                            f.write(diff + "\n")
                        with open(".summary/changes.diff", "w") as f:
                            f.write(diff + "\n")
                    else:
                        shutil.copy(
                            regelum.main.metadata["common_dir"] + "/changes.diff",
                            ".summary/changes.diff",
                        )
            if forbidden:
                raise Exception(
                    "Running experiments without committing is disallowed. Please, commit your changes."
                )
        except git.exc.InvalidGitRepositoryError as err:
            commit_hash = None
            repo = None
            if (
                "disallow_uncommitted" in regelum.main.cfg
                and regelum.main.cfg.disallow_uncommitted
                and not is_in_debug_mode()
            ):
                raise Exception(
                    "Running experiments without committing is disallowed. Please, commit your changes."
                ) from err
        return repo, commit_hash

    @disable_in_jupyter
    def on_launch(self):
        cfg = regelum.main.config
        metadata = regelum.main.metadata
        report = metadata["report"]
        start = time.time()
        os.mkdir(".summary")
        name = metadata["config_path"].split("/")[-1].split(".")[0]
        cfg.treemap(root=name).write_html("SUMMARY.html")
        with open("SUMMARY.html", "r") as f:
            html = f.read()
        repo, commit_hash = (
            self.__monitor_git()
            if not regelum.main.metadata["no_git"]
            else (None, None)
        )
        cfg_hash = hex(hash(cfg))
        html = html.replace("<body>", f"<title>{name} {cfg_hash}</title><body>")
        overrides_table = ""
        with open(".rehydra/rehydra.yaml", "r") as f:
            content = f.read()
            content = content[content.find("task:") :]
            content = content[: content.find("job:")]
            content = content.replace("-", "").split()[1:]
            if content[0] != "[]":
                with report() as r:
                    r["overrides"] = {}
                    for line in content:
                        field, value = line.split("=")
                        overrides_table += f'<tr><td><font face="Courier New">{field}</font></td> <td><font face="Courier New"> = </font></td>  <td><font face="Courier New">{value}</font></td> </tr>\n'
                        r["overrides"][field] = value
                    r["overrides_html"] = f"<table>{overrides_table}</table>"
            else:
                with report() as r:
                    r["overrides"] = {}
                    r["overrides_html"] = ""
        html = html.replace(
            "<head>",
            """
                            <head>
                              <link rel="mask-icon" type="image/x-icon" href="https://cpwebassets.codepen.io/assets/favicon/logo-pin-b4b4269c16397ad2f0f7a01bcdf513a1994f4c94b8af2f191c09eb0d601762b1.svg" color="#111" />  
                              <link rel='stylesheet' href='https://cdnjs.cloudflare.com/ajax/libs/prism/1.22.0/themes/prism-tomorrow.min.css'>
                              
                            <style>
                            *,
                            *:before,
                            *:after {
                              box-sizing: border-box;
                            }
                            
                            pre[class*="language-"] {
                              position: relative;
                              overflow: auto;
                            
                              /* make space  */
                              margin: 5px 0;
                              padding: 1.75rem 0 1.75rem 1rem;
                              border-radius: 10px;
                            }
                            
                            pre[class*="language-"] button {
                              position: absolute;
                              top: 5px;
                              right: 5px;
                            
                              font-size: 0.9rem;
                              padding: 0.15rem;
                              background-color: #828282;
                            
                              border: ridge 1px #7b7b7c;
                              border-radius: 5px;
                              text-shadow: #c4c4c4 0 0 2px;
                            }
                            
                            pre[class*="language-"] button:hover {
                              cursor: pointer;
                              background-color: #bcbabb;
                            }
                            
                            main {
                              display: grid;
                              max-width: 600px;
                              margin: 20px auto;
                            }
                            
                            </style>
                            
                              <script>
                              window.console = window.console || function(t) {};
                            </script>
                              
                              <script>
                              if (document.location.search.match(/type=embed/gi)) {
                                window.parent.postMessage("resize", "*");
                              }
                            </script>
                            """,
        )
        html = html.replace(
            "<body>",
            f"""
                            <body>
                            <div>
                            <table style="margin-left: auto; margin-right: 0;">
                            <tbody>
                            {overrides_table}
                            </tbody>
                            </table>
                            </div>
                            </div>
            """,
        )
        html = html.replace(
            "<body>",
            f"""
                            <body>
                            <div style="display: grid; grid-template-columns: 1fr 1fr;">
                            <div>
                            <table>
                            <tbody>
                            <tr><td>Config hash:  </td> <td>{cfg_hash}</td></tr>
                            <tr><td>Commit hash: </td><td>{commit_hash} </td></tr>
                            <tr><td>Date and time: </td><td>{datetime.datetime.now().strftime("%d.%m.%Y %H:%M:%S")} </td></tr>
                            <tr><td>Script path: </td><td>{metadata["script_path"]} </td></tr>
                            <tr><td>Config path: </td><td>{metadata["config_path"]}</td></tr>
                            </tbody>
                            </table>
                            </div>
            """,
        )
        installed_packages = pkg_resources.working_set
        installed_packages_list = sorted(
            [(i.key, i.version) for i in installed_packages]
        )
        packages_table = []
        for package, version in installed_packages_list:
            if package in [
                "matplotlib",
                "numpy",
                "scipy",
                "omegaconf",
                "hydra-core",
                "hydra-joblib-launcher",
                "pandas",
                "gpytorch",
                "torch",
                "casadi",
                "dill",
                "plotly",
                "regelum",
            ]:
                packages_table = [
                    f'<tr><td><font face="Courier New" color="red">{package}</font></td> <td><font face="Courier New" color="red"> == </font></td>  <td><font face="Courier New" color="red">{version}</font></td> </tr>\n'
                ] + packages_table
            else:
                packages_table.append(
                    f'<tr><td><font face="Courier New">{package}</font></td> <td><font face="Courier New"> == </font></td>  <td><font face="Courier New">{version}</font></td> </tr>\n'
                )

        html = html.replace(
            "</body>",
            f"""
                            <br>
                            <details>
                            <summary> Environment details </summary>
                            <table>
                            <tr style="vertical-align:top"> 
                            <td> <table>  {"".join(packages_table[:len(packages_table) // 4 ])}  </table> </td>
                            <td> <table>  {"".join(packages_table[len(packages_table) // 4 :2 * len(packages_table) // 4])}  </table> </td>
                            <td> <table>  {"".join(packages_table[2 *  len(packages_table) // 4: 3 * len(packages_table) // 4])}  </table> </td>
                            <td> <table>  {"".join(packages_table[3 *  len(packages_table) // 4:])}  </table> </td>
                            </tr>
                            </table>
                            </details>
                             </body>
            """,
        )
        git_fragment = (
            f"""git checkout {commit_hash.replace(' <font color="red">(uncommitted/unstaged changes)</font>',  chr(10) + f'patch -p1 < {os.path.abspath(".summary/changes.diff")}')}"""
            + chr(10)
            if commit_hash
            else ""
        )
        html = html.replace(
            "</body>",
            f"""
              <details>
              <summary>Snippets</summary>
              <main>
              <p>Extract callbacks:</p>
              <pre><code class="language-python">import dill, os, sys
os.chdir("{metadata["initial_working_directory"]}")
sys.path[:0] = {metadata["initial_pythonpath"].split(":")}
with open("{os.path.abspath(".")}/callbacks.dill", "rb") as f:
    callbacks = dill.load(f)</code></pre>"""
            + (
                f"""
              <p>Reproduce experiment:</p>
              <pre><code class="language-bash">cd {repo.working_tree_dir}
git reset
git restore .
git clean -f
{git_fragment}cd {metadata["initial_working_directory"]}
export PYTHONPATH="{metadata["initial_pythonpath"]}"
python3 {metadata["script_path"]} {" ".join(content if content[0] != "[]" else [])} {" ".join(list(filter(lambda x: "--" in x and "multirun" not in x, sys.argv)))} </code></pre>
            """
                if repo is not None
                else ""
            )
            + """</main>
                <script src="https://cpwebassets.codepen.io/assets/common/stopExecutionOnTimeout-2c7831bb44f98c1391d6a4ffda0e1fd302503391ca806e7fcc7b9b87197aec26.js"></script>
            
              <script src='https://cdnjs.cloudflare.com/ajax/libs/prism/1.26.0/prism.min.js'></script>
              <script src="https://cdnjs.cloudflare.com/ajax/libs/prism/1.26.0/components/prism-python.min.js"></script>
              <script src="https://cdnjs.cloudflare.com/ajax/libs/prism/1.26.0/components/prism-bash.min.js"></script>
                  <script>
            const copyButtonLabel = "Copy Code";
            
            // use a class selector if available
            let blocks = document.querySelectorAll("pre");
            
            blocks.forEach(block => {
              // only add button if browser supports Clipboard API
              if (navigator.clipboard) {
                let button = document.createElement("button");
            
                button.innerText = copyButtonLabel;
                block.appendChild(button);
            
                button.addEventListener("click", async () => {
                  await copyCode(block, button);
                });
              }
            });
            
            async function copyCode(block, button) {
              let code = block.querySelector("code");
              let text = code.innerText;
            
              await navigator.clipboard.writeText(text);
            
              // visual feedback that task is completed
              button.innerText = "Code Copied";
            
              setTimeout(() => {
                button.innerText = copyButtonLabel;
              }, 700);
            }
            //# sourceURL=pen.js
                </script>
            </details>
            </body>
            """,
        )
        with open("SUMMARY.html", "w") as f:
            f.write(html)
        self.log(
            f"Saved summary to {os.path.abspath('SUMMARY.html')}. ({int(1000 * (time.time() - start))}ms)"
        )

    def on_termination(self, res):
        with open("SUMMARY.html", "r") as f:
            html = f.read()
        table_lines = ""
        images = []
        directory = os.fsencode("gfx")

        for file in os.listdir(directory):
            filename = os.fsdecode(file)
            images.append(
                f'<img src="gfx/{filename}" style="object-fit: cover; width: 100%; max-height: 100%;">'
            )
        for image in images:
            table_lines += f"<div>{image}</div>\n"
        html = html.replace(
            "</body>",
            f"""
                            <br>
                            <div style="display: grid; grid-template-columns: 1fr 1fr;">
                            {table_lines}
                            </div>
                             </body>
                             """,
        )
        with open("SUMMARY.html", "w") as f:
            f.write(html)


@trigger
class StateTracker(Callback):
    """Records the state of the simulated system into `self.system_state`.

    Useful for animations that visualize motion of dynamical systems.
    """

    def is_target_event(self, obj, method, output, triggers):
        return (
            isinstance(obj, regelum.scenario.Scenario)
            and method == "post_compute_action"
        )

    def is_done_collecting(self):
        return hasattr(self, "system_state")

    def on_function_call(self, obj, method, output):
        self.system_state = obj.state
        self.system_state = self.system_state.reshape(self.system_state.size)
        self.state_naming = obj.simulator.system.state_naming


@trigger
class TimeTracker(Callback):
    """Records the state of the simulated system into `self.system_state`.

    Useful for animations that visualize motion of dynamical systems.
    """

    def is_target_event(self, obj, method, output, triggers):
        return (
            isinstance(obj, regelum.scenario.Scenario)
            and method == "post_compute_action"
        )

    def is_done_collecting(self):
        return hasattr(self, "time")

    def on_function_call(self, obj, method, output):
        self.time = output["time"]


@trigger
class ObservationTracker(Callback):
    """Records the state of the simulated system into `self.system_state`.

    Useful for animations that visualize motion of dynamical systems.
    """

    def is_target_event(self, obj, method, output, triggers):
        return (
            isinstance(obj, regelum.scenario.Scenario)
            and method == "post_compute_action"
        )

    def is_done_collecting(self):
        return hasattr(self, "observation")

    def on_function_call(self, obj, method, output):
        self.observation = output["observation"][0]
        self.observation_naming = obj.simulator.system.observation_naming


@trigger
class ObjectiveTracker(Callback):
    """Records the state of the simulated system into `self.system_state`.

    Useful for animations that visualize motion of dynamical systems.
    """

    def is_target_event(self, obj, method, output, triggers):
        return (
            isinstance(obj, regelum.scenario.Scenario)
            and method == "post_compute_action"
        )

    def is_done_collecting(self):
        return hasattr(self, "objective")

    def on_function_call(self, obj, method, output):
        self.running_objective = output["running_objective"]
        self.value = output["current_value"]
        self.objective = np.array([self.value, self.running_objective])
        self.objective_naming = ["Value", "Running objective"]


@trigger
class ValueTracker(Callback):
    """Records the state of the simulated system into `self.system_state`.

    Useful for animations that visualize motion of dynamical systems.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.counter = None

    def is_target_event(self, obj, method, output, triggers):
        return isinstance(obj, regelum.scenario.Scenario) and (
            method == "reload_scenario"
        )

    def is_done_collecting(self):
        return hasattr(self, "score")

    def on_function_call(self, obj, method, output):
        self.score = obj.recent_value


@trigger
class ActionTracker(Callback):
    """Records the state of the simulated system into `self.system_state`.

    Useful for animations that visualize motion of dynamical systems.
    """

    def is_target_event(self, obj, method, output, triggers):
        return (
            isinstance(obj, regelum.scenario.Scenario)
            and method == "post_compute_action"
        )

    def is_done_collecting(self):
        return hasattr(self, "action")

    def on_function_call(self, obj, method, output):
        self.action = output["action"][0]
        self.action_naming = obj.simulator.system._inputs_naming


class HistoricalCallback(Callback, ABC):
    """Callback (base) responsible for recording various temporal data."""

    def __init__(self, *args, **kwargs):
        """Initialize an instance of HistoricalCallback."""
        super().__init__(*args, **kwargs)
        self.xlabel = "Time"
        self.ylabel = "Value"
        self.data = pd.DataFrame()

        self.save_directory = Path(f".callbacks/{self.__class__.__name__}").resolve()

    def get_save_directory(self):
        return self.save_directory

    def on_launch(self):
        os.mkdir(self.get_save_directory())

    def add_datum(self, datum: dict):
        if self.data.empty:
            self.data = pd.DataFrame({key: [value] for key, value in datum.items()})
        else:
            self.data.loc[len(self.data)] = datum

    @disable_in_jupyter
    def dump_data(self, identifier):
        if not self.data.empty:
            self.data.to_hdf(
                f".callbacks/{self.__class__.__name__}/{identifier}.h5", key="data"
            )

    def load_data(self, idx=None):
        dirs = sorted(Path(self.get_save_directory()).iterdir())

        if len(dirs) > 0:
            if idx is None:
                return pd.concat(
                    [pd.read_hdf(path) for path in dirs],
                    axis=0,
                )
            else:
                if idx == "last":
                    idx = len(dirs)
                assert idx >= 1, "Indices should be no smaller than 1."
                assert idx <= len(dirs), f"Only {len(dirs)} files were stored."
                return pd.read_hdf(dirs[idx - 1])
        else:
            return pd.DataFrame()

    def insert_column_left(self, column_name, values):
        if not self.data.empty:
            self.data.insert(0, column_name, values)

    @disable_in_jupyter
    def dump_and_clear_data(self, identifier):
        self.dump_data(identifier)
        self.clear_recent_data()

    def clear_recent_data(self):
        self.data = pd.DataFrame()

    def plot(self, name=None):
        if regelum.main.is_clear_matplotlib_cache_in_callbacks:
            plt.clf()
            plt.cla()
            plt.close()
        if not name:
            name = self.__class__.__name__
        res = self.data.plot()
        plt.xlabel(self.xlabel)
        plt.ylabel(self.ylabel)
        plt.title(name)
        plt.grid()
        # plt.xticks(range(1, len(self.data) + 1))
        return res.figure

    def save_plot(self, name=None):
        if not name:
            name = self.__class__.__name__
        self.plot(name=name)
        plt.savefig(f"gfx/{name}.svg")
        mlflow.log_artifact(f"gfx/{name}.svg", "gfx")

    def plot_gui(self):
        self.data = self.load_data(idx="last")
        if not self.data.empty:
            return self.plot(name=self.__class__.__name__)
        else:
            return None


class CalfCallback(HistoricalCallback):
    """Callback that records various diagnostic data during experiments with CALF-based methods."""

    def __init__(self, *args, **kwargs):
        """Initialize an instance of CalfCallback."""
        super().__init__(*args, **kwargs)
        self.cooldown = 1.0
        self.time = 0.0

    def is_target_event(self, obj, method, output):
        return (
            isinstance(obj, regelum.scenario.Scenario)
            and method == "compute_action"
            and "calf" in obj.critic.__class__.__name__.lower()
        )

    def perform(self, obj, method, output):
        current_CALF = obj.critic(
            obj.critic.observation_last_good,
            use_stored_weights=True,
        )
        self.log(
            f"current CALF value:{current_CALF}, decay_rate:{obj.critic.safe_decay_rate}, observation: {obj.data_buffer.sample_last(keys=['observation'], dtype=np.array)['observation']}"
        )
        is_calf = (
            obj.critic.opt_status == regelum.optimizable.optimizers.OptStatus.success
            and obj.policy.opt_status
            == regelum.optimizable.optimizers.OptStatus.success
        )
        if not self.data.empty:
            prev_CALF = self.data["J_hat"].iloc[-1]
        else:
            prev_CALF = current_CALF

        delta_CALF = prev_CALF - current_CALF
        self.add_datum(
            {
                "time": obj.data_buffer.get_latest("time"),
                "J_hat": (
                    current_CALF[0]
                    if isinstance(current_CALF, (np.ndarray, torch.Tensor))
                    else current_CALF.full()[0]
                ),
                "is_CALF": is_calf,
                "delta": (
                    delta_CALF[0]
                    if isinstance(delta_CALF, (np.ndarray, torch.Tensor))
                    else delta_CALF.full()[0]
                ),
            }
        )

    def on_episode_done(
        self,
        scenario,
        episode_number,
        episodes_total,
        iteration_number,
        iterations_total,
    ):
        identifier = f"CALF_diagnostics_on_episode_{str(iteration_number).zfill(5)}"
        if not self.data.empty:
            self.save_plot(identifier)
            self.insert_column_left("iteration", iteration_number)
            self.dump_and_clear_data(identifier)

    def plot(self, name=None):
        if regelum.main.is_clear_matplotlib_cache_in_callbacks:
            plt.clf()
            plt.cla()
            plt.close()
        if not self.data.empty:
            fig = plt.figure(figsize=(10, 10))

            ax_calf, ax_switch, ax_delta = ax_array = fig.subplots(1, 3)
            ax_calf.plot(self.data["time"], self.data["J_hat"], label="CALF")
            ax_switch.plot(self.data["time"], self.data["is_CALF"], label="CALF on")
            ax_delta.plot(self.data["time"], self.data["delta"], label="delta decay")

            for ax in ax_array:
                ax.set_xlabel("Time [s]")
                ax.grid()
                ax.legend()

            ax_calf.set_ylabel("J_hat")
            ax_switch.set_ylabel("Is CALF on")
            ax_delta.set_ylabel("Decay size")

            plt.legend()
            return fig


class CALFWeightsCallback(HistoricalCallback):
    """Callback that records the relevant model weights for CALF-based methods."""

    def __init__(self, *args, **kwargs):
        """Initialize an instance of CALFWeightsCallback."""
        super().__init__(*args, **kwargs)
        self.cooldown = 0.0
        self.time = 0.0

    def is_target_event(self, obj, method, output):
        return (
            isinstance(obj, regelum.scenario.Scenario)
            and method == "compute_action"
            and "calf" in obj.critic.__class__.__name__.lower()
        )

    def perform(self, obj, method, output):
        try:
            datum = {
                **{
                    "time": obj.data_buffer.get_latest("time"),
                },
                **{
                    f"weight_{i + 1}": weight
                    for i, weight in enumerate(
                        obj.critic.model.weights
                        if isinstance(
                            obj.critic.model.weights, (np.ndarray, torch.Tensor)
                        )
                        else obj.critic.model.weights.full().reshape(-1)
                    )
                },
            }

            # print(datum["time"], obj.critic.model.weights)
            self.add_datum(datum)
        finally:
            pass

    def on_episode_done(
        self,
        scenario,
        episode_number,
        episodes_total,
        iteration_number,
        iterations_total,
    ):
        identifier = f"CALF_weights_during_episode_{str(iteration_number).zfill(5)}"
        if not self.data.empty:
            self.save_plot(identifier)
            self.insert_column_left("episode", iteration_number)
            self.dump_and_clear_data(identifier)

    def plot(self, name=None):
        if not self.data.empty:
            if regelum.main.is_clear_matplotlib_cache_in_callbacks:
                plt.clf()
                plt.cla()
                plt.close()
            if not name:
                name = self.__class__.__name__
            res = self.data.set_index("time").plot(
                subplots=True, grid=True, xlabel="time", title=name
            )
            return res[0].figure


def method_callback(method_name, class_name=None, log_level="debug"):
    """Create a callback class that logs the output of a specific method of a class or any class.

    Args:
        method_name (str): Name of the method to log output for.
        class_name (str or class): (Optional) Name of the class the
            method belongs to. If not specified, the callback will log
            the output for the method of any class.
        log_level (str): (Optional) The level of logging to use. Default
            is "debug".
    """
    if class_name is not None:
        class_name = (
            class_name.__name__ if not isinstance(class_name, str) else class_name
        )

    class MethodCallback(Callback):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def is_target_event(self, obj, method, output, triggers):
            return method == method_name and class_name in [
                None,
                obj.__class__.__name__,
            ]

        def on_function_call(self, obj, method, output):
            self.log(
                f"Method '{method}' of class '{obj.__class__.__name__}' returned {output}"
            )

    return MethodCallback


class ScenarioStepLogger(Callback):
    """A callback which allows to log every step of simulation in a scenario."""

    cooldown = 1.0

    def is_target_event(self, obj, method, output, triggers):
        return (
            isinstance(obj, regelum.scenario.Scenario)
            and method == "post_compute_action"
        )

    def on_function_call(self, obj, method: str, output: Dict[str, Any]):
        with np.printoptions(precision=2, suppress=True):
            self.log(
                f"runn. objective: {output['running_objective']:.2f}, "
                f"state est.: {output['estimated_state'][0]}, "
                f"observation: {output['observation'][0]}, "
                f"action: {output['action'][0]}, "
                f"value: {output['current_value']:.4f}, "
                f"time: {output['time']:.4f} ({100 * output['time']/obj.simulator.time_final:.1f}%), "
                f"episode: {int(output['episode_id'])}/{obj.N_episodes}, "
                f"iteration: {int(output['iteration_id'])}/{obj.N_iterations}"
            )


class SaveProgressCallback(Callback):
    """Callback responsible for regularly saving data collected by other callbacks to the hard-drive."""

    once_in = 1

    def on_launch(self):
        with regelum.main.metadata["report"]() as r:
            r["episode_current"] = 0

    def is_target_event(self, obj, method, output, triggers):
        return False

    def on_function_call(self, obj, method, output):
        pass

    @disable_in_jupyter
    def on_episode_done(
        self,
        scenario,
        episode_number,
        episodes_total,
        iteration_number,
        iterations_total,
    ):
        start = time.time()
        if episode_number % self.once_in:
            return
        filename = "callbacks.dill"
        with open(filename, "wb") as f:
            dill.dump(regelum.main.callbacks, f)
        self.log(
            f"Saved callbacks to {os.path.abspath(filename)}. ({int(1000 * (time.time() - start))}ms)"
        )
        if scenario is not None:
            with regelum.main.metadata["report"]() as r:
                r["episode_current"] = episode_number
                if "episodes_total" not in r:
                    r["episode_total"] = episodes_total
                r["elapsed_relative"] = 1.0

    def on_termination(self, res):
        if isinstance(res, Exception):
            self.on_episode_done(None, 0, None, None, None)


class HistoricalDataCallback(HistoricalCallback):
    """A callback which allows to store desired data collected among different runs inside multirun execution runtime."""

    def __init__(self, *args, **kwargs):
        """Initialize an instance of HistoricalDataCallback."""
        super().__init__(*args, **kwargs)
        self.cooldown = 0.0

        self.observation_components_naming = None
        self.action_components_naming = None
        self.state_components_naming = None

    def is_target_event(self, obj, method, output, triggers):
        return isinstance(obj, regelum.scenario.Scenario) and (
            method == "post_compute_action" or method == "dump_data_buffer"
        )

    def on_function_call(self, obj, method, output):
        if self.observation_components_naming is None:
            self.observation_components_naming = (
                [
                    f"observation_{i + 1}"
                    for i in range(obj.simulator.system.dim_observation)
                ]
                if obj.simulator.system.observation_naming is None
                else obj.simulator.system.observation_naming
            )

        if self.action_components_naming is None:
            self.action_components_naming = (
                [f"action_{i + 1}" for i in range(obj.simulator.system.dim_inputs)]
                if obj.simulator.system.inputs_naming is None
                else obj.simulator.system.inputs_naming
            )

        if self.state_components_naming is None:
            self.state_components_naming = (
                [f"state_{i + 1}" for i in range(obj.simulator.system.dim_state)]
                if obj.simulator.system.state_naming is None
                else obj.simulator.system.state_naming
            )

        if method == "post_compute_action":
            self.add_datum(
                {
                    **{
                        "time": output["time"],
                        "running_objective": output["running_objective"],
                        "current_value": output["current_value"],
                        "episode_id": output["episode_id"],
                        "iteration_id": output["iteration_id"],
                    },
                    **dict(zip(self.action_components_naming, output["action"][0])),
                    **dict(
                        zip(self.state_components_naming, output["estimated_state"][0])
                    ),
                    # **dict(
                    #     zip(self.state_components_naming, output["estimated_state"][0])
                    # ),
                }
            )
        elif method == "dump_data_buffer":
            _, data_buffer = output
            self.data = pd.concat(
                [
                    data_buffer.to_pandas(
                        keys={
                            "time": float,
                            "running_objective": float,
                            "current_value": float,
                            "episode_id": int,
                            "iteration_id": int,
                        }
                    )
                ]
                + [
                    pd.DataFrame(
                        columns=columns,
                        data=np.array(
                            data_buffer.to_pandas([key]).values.tolist(),
                            dtype=float,
                        ).squeeze(),
                    )
                    for columns, key in [
                        (self.action_components_naming, "action"),
                        (self.state_components_naming, "estimated_state"),
                        # (self.state_components_naming, "estimated_state"),
                    ]
                ],
                axis=1,
            )

    @disable_in_jupyter
    def on_episode_done(
        self,
        scenario,
        episode_number,
        episodes_total,
        iteration_number,
        iterations_total,
    ):
        if episodes_total == 1:
            identifier = f"observations_actions_it_{str(iteration_number).zfill(5)}"
        else:
            identifier = f"observations_actions_it_{str(iteration_number).zfill(5)}_ep_{str(episode_number).zfill(5)}"
        self.save_plot(identifier)
        self.dump_and_clear_data(identifier)

    def plot(self, name=None):
        if regelum.main.is_clear_matplotlib_cache_in_callbacks:
            plt.clf()
            plt.cla()
            plt.close()

        if not name:
            name = self.__class__.__name__

        axes = (
            self.data[
                self.state_components_naming + self.action_components_naming + ["time"]
            ]
            .set_index("time")
            .plot(subplots=True, grid=True, xlabel="time", title=name, legend=False)
        )
        for ax, label in zip(
            axes, self.state_components_naming + self.action_components_naming
        ):
            ax.set_ylabel(label)

        return axes[0].figure


class ObjectiveSaver(HistoricalCallback):
    """A callback which allows to store objective values during execution runtime."""

    def __init__(self, *args, **kwargs):
        """Initialize an instance of PolicyGradientObjectiveSaverCallback."""
        super().__init__(*args, **kwargs)

        self.cooldown = 0.0
        self.iteration_number = 1

    def is_target_event(self, obj, method, output, triggers):
        return (
            isinstance(obj, regelum.scenario.RLScenario) and (method == "pre_optimize")
        ) or (
            isinstance(obj, regelum.optimizable.Optimizable)
            and (method == "post_epoch" or method == "post_optimize")
        )

    def on_function_call(self, obj, method, output):
        if isinstance(obj, regelum.scenario.RLScenario) and (method == "pre_optimize"):
            which, event, time, episode_counter, iteration_counter = output
            time = float(time)
            if which == "Critic":
                self.key = f"B. {which} objective. "
            else:
                self.key = f"C. {which} objective. "

            if event == Event.compute_action:
                if time is None:
                    raise ValueError("Time should be passed if one uses compute action")
                self.key += f"Time {str(round(time, 4))}. Ep {str(episode_counter).zfill(5)}. It {str(iteration_counter).zfill(5)}"
            elif event == Event.reset_episode:
                self.key += f"Ep {str(episode_counter).zfill(5)}. It {str(iteration_counter).zfill(5)}"
            elif event == Event.reset_iteration:
                self.key += f"It {str(iteration_counter).zfill(5)}"

        if isinstance(obj, regelum.optimizable.Optimizable) and (
            method == "post_epoch"
        ):
            epoch_idx, objective_values = output
            if len(objective_values) > 0:
                mlflow.log_metric(
                    self.key,
                    np.mean(objective_values),
                    step=epoch_idx,
                )
                self.add_datum(
                    {"epoch_idx": epoch_idx, "objective": np.mean(objective_values)}
                )

        if isinstance(obj, regelum.optimizable.Optimizable) and (
            method == "post_optimize"
        ):
            self.dump_and_clear_data(self.key)


class CriticObjectiveSaver(ObjectiveSaver):
    """A callback which allows to store critic objective values during execution runtime."""

    def is_target_event(self, obj, method, output, triggers):
        return (
            isinstance(obj, regelum.scenario.RLScenario) and (method == "pre_optimize")
        ) or (
            isinstance(obj, regelum.critic.Critic)
            and (method == "post_epoch" or method == "post_optimize")
        )


class PolicyObjectiveSaver(ObjectiveSaver):
    """A callback which allows to store critic objective values during execution runtime."""

    def is_target_event(self, obj, method, output, triggers):
        return (
            isinstance(obj, regelum.scenario.RLScenario) and (method == "pre_optimize")
        ) or (
            isinstance(obj, regelum.policy.Policy)
            and (method == "post_epoch" or method == "post_optimize")
        )


class ValueCallback(HistoricalCallback):
    """Callback that regularly logs value."""

    def __init__(self, *args, **kwargs):
        """Initialize an instance of TotalObjectiveCallback."""
        super().__init__(*args, **kwargs)
        self.cache = pd.DataFrame()
        self.xlabel = "Episode"
        self.ylabel = "value"
        self.values = []
        self.mean_iteration_values = []

    def is_target_event(self, obj, method, output, triggers):
        return False

    @disable_in_jupyter
    def on_episode_done(
        self,
        scenario,
        episode_number,
        episodes_total,
        iteration_number,
        iterations_total,
    ):
        self.add_datum(
            {
                "episode": len(self.data) + 1,
                "objective": scenario.recent_value,
            }
        )
        self.values.append(scenario.recent_value)
        self.log(
            f"Final value of episode {self.data.iloc[-1]['episode']} is {round(self.data.iloc[-1]['objective'], 2)}"
        )
        if not self.is_jupyter:
            self.dump_data(
                f"Total_Objectives_in_iteration_{str(iteration_number).zfill(5)}"
            )
            mlflow.log_metric(
                f"C. values in iteration {str(iteration_number).zfill(5)}",
                scenario.recent_value,
                step=len(self.values),
            )

            self.save_plot(
                f"Total_Objectives_in_iteration_{str(iteration_number).zfill(5)}"
            )

        if episode_number == episodes_total:
            if not self.is_jupyter:
                mlflow.log_metric(
                    "A. Avg value",
                    np.mean(self.values),
                    step=iteration_number,
                )
                self.mean_iteration_values.append(np.mean(self.values))
                self.values = []

                mlflow.log_metric(
                    "A. Cum max value",
                    np.max(self.mean_iteration_values),
                    step=iteration_number,
                )
                mlflow.log_metric(
                    "A. Cum min value",
                    np.min(self.mean_iteration_values),
                    step=iteration_number,
                )
            else:
                self.log(f"Avg value: {np.mean(self.values)}")

            self.clear_recent_data()

    def on_function_call(self, obj, method, output):
        pass

    def load_data(self, idx=None):
        return super().load_data(idx=1)

    def plot(self, name=None):
        if regelum.main.is_clear_matplotlib_cache_in_callbacks:
            plt.clf()
            plt.cla()
            plt.close()
        if not name:
            name = self.__class__.__name__

        res = self.data.set_index("episode").plot()
        plt.xlabel(self.xlabel)
        plt.ylabel(self.ylabel)
        plt.title(name)
        plt.grid()
        plt.xticks(range(1, len(self.data) + 1))
        return res.figure


class TimeRemainingCallback(Callback):
    """Callback that logs an estimate of the time that remains till the end of the simulation."""

    def __init__(self, *args, **kwargs):
        """Initialize an instance of TimeRemainingCallback."""
        super().__init__(*args, **kwargs)
        self.time_episode = []

    def is_target_event(self, obj, method, output, triggers):
        return False

    def on_function_call(self, obj, method, output):
        pass

    def on_episode_done(
        self,
        scenario,
        episode_number,
        episodes_total,
        iteration_number,
        iterations_total,
    ):
        self.time_episode.append(time.time())
        if len(self.time_episode) > 3:
            average_interval = 0
            previous = self.time_episode[-1]
            for current in self.time_episode[-2:-12:-1]:
                average_interval -= current - previous
                previous = current
            average_interval /= len(self.time_episode[-2:-12:-1])
            episodes_left = (
                episodes_total
                - episode_number
                + (iterations_total - iteration_number) * episodes_total
            )
            td = datetime.timedelta(seconds=int(episodes_left * average_interval))
            remaining = f" Estimated remaining time: {str(td)}."
        else:
            remaining = ""
        self.log(
            f"Completed episode {episode_number}/{episodes_total} ({100*episode_number/episodes_total:.1f}%). Iteration: {iteration_number}/{iterations_total}."
            + remaining
        )


from .animation import *



"""Tools for creating and retrieving user settings."""

import stat
import subprocess
from pathlib import Path
import os
from cues import Select
from omegaconf import OmegaConf
import cues
from rich import print

from regelum.__internal.cues_fix import RegelumSurvey

user_data_dir = Path(__file__).parent.parent
config_file = user_data_dir / "user_settings.yaml"

default_user_settings = {
    "MLFLOW_TRACKING_URI": None,
    "DISALLOW_UNCOMMITTED": True,
    "JOBS": 1,
    "IS_CONTRIBUTOR": False,
    "FULL_NAME": None,
    "E_MAIL": None,
    "JOBS_TESTING": 1,
}


def get_user_settings():
    if not os.path.exists(config_file):
        configure_user_settings()
    return OmegaConf.load(config_file)


def choice(*options, message=None):
    assert message is not None
    return Select("name", message, options).send()["name"]


def ok(message):
    return choice("Ok", message=message)


def form(*fields, message=None):
    assert message is not None
    fields = [{"name": i, "message": field} for i, field in enumerate(fields)]
    return cues.Form("name", message, fields).send()["name"]


def configure_user_settings():
    user_settings = default_user_settings.copy()
    if not os.path.exists(user_data_dir):
        os.mkdir(user_data_dir)
    print(
        "Welcome to [bold magenta]regelum[/bold magenta]! Before we proceed let's quickly configure your setup."
    )
    name = "cores"
    message = "Please select the number of cores that regelum will use by default:"
    scale = range(1, os.cpu_count() + 1)
    fields = [{"name": "cores", "message": "Cores"}]
    cue = RegelumSurvey(name, message, scale, fields)
    user_settings["JOBS"] = cue.send()["cores"]["cores"]
    ok(
        "This choice can always be overriden by specifying [bold magenta]--jobs=[/bold magenta] in command line arguments."
    )
    menu = [
        "No, just keep the artifacts in ./regelum_data/mlflow.",
        "Yes, let me set the URI.",
        'I don\'t even know what "mlflow" is!',
    ]
    msg = "Regelum uses [bold magenta]mlflow[/bold magenta] to store experiment data. By default a subfolder is created in the current working directory to store mlflow artifacts. Would you like to set a custom mlflow URI?"
    ans = choice(*menu, message=msg)
    if ans == menu[0]:
        user_settings["MLFLOW_TRACKING_URI"] = None
    elif ans == menu[1]:
        ans = form(
            "Tracking URI", message="Specify the URI to your mlflow data server."
        )[0]
        user_settings["MLFLOW_TRACKING_URI"] = ans
    else:
        user_settings["MLFLOW_TRACKING_URI"] = None
        ok(
            "Basically, [bold magenta]mlflow[/bold magenta] is a browser user interface that let's you see your [bold magenta]graphs[/bold magenta], [bold magenta]metrics[/bold magenta] and [bold magenta]animations[/bold magenta]. You also get to easilly compare the outcomes of different experiments. It might not sound like much, but, in fact, it is almost impossible to get by without a tool like this."
        )
        ok(
            "This 'tracking uri' thing is what lets you specify your own mlflow server, which can be useful if you have a group of people working on the same collection of experiments. Since you probably don't have a server like this yet, we'll just store your mlflow experiment data in [bold magenta]./regelum_data/mlflow[/bold magenta]."
        )
        ok(
            "To view your experiment data via mlflow, all you need to do is:\n [bold grey]cd ./regelum_data[/bold grey]\n [bold grey]mlflow ui[/bold grey]\n Then just click the link that shows up in the logs."
        )
    ok(
        "By the way, don't forget to add [bold magenta]regelum_data/[/bold magenta] to your [bold magenta].gitignore[/bold magenta]! Regelum will always create one of these folders in your current working directory."
    )
    ans = choice(
        "Yes, prevent me from running experiments if I have uncommitted changes.",
        "No, let me run regelum even when my repo has uncommitted changes.",
        message="By default regelum, [bold magenta]will not start unless all changes in your repo are committed[/bold magenta]. This helps to ensure reproducibility. If you're doing serious research, it will help you a lot, but if not, then it might be really annoying. Would you like to keep this feature enabled?",
    )
    user_settings["DISALLOW_UNCOMMITTED"] = "Yes" in ans
    name, email = form(
        "Full name",
        "E-mail",
        message="At the end of each experiment regelum generates a [bold magenta]summary[/bold magenta] that outlines the setup, the outcomes and steps to reproduce the results. When working in a team it is useful to have credentials of the author of the experiment present in this summary. Please provide your [bold magenta]e-mail[/bold magenta] and [bold magenta]e-mail[/bold magenta] address or leave the lines blank.",
    )
    user_settings["FULL_NAME"] = name if name else None
    user_settings["E_EMAIL"] = email if email else None
    script_dev = os.path.abspath(__file__ + "/../../scripts/developer-setup.sh")
    user_settings["IS_CONTRIBUTOR"] = False
    if os.path.exists(script_dev):
        ans = choice(
            "No, I'm only a user of regelum and I do not intend to modify it.",
            "Yes, I am a developer of regelum.",
            message="Last question. Are you a contributor?",
        )
        if "Yes" in ans:
            ok("In that case, let me install some additional dependencies...")
            try:
                st = os.stat(script_dev)
                os.chmod(script_dev, st.st_mode | stat.S_IEXEC)
                subprocess.run(script_dev, shell=True, check=True)
            except subprocess.CalledProcessError as e:
                print(e)
                ok(
                    f"It would seem that something went wrong when setting up your environment for contributing to regelum. Before you can start your project you need to make sure that all the necessary pre-requisites are installed. Please try to determine, what went wrong when running {script_dev}."
                )
                return
            user_settings["IS_CONTRIBUTOR"] = True
            name = "cores"
            message = (
                "Now, please specify how many cores are to be used during testing:"
            )
            scale = range(1, os.cpu_count() + 1)
            fields = [{"name": "cores", "message": "Cores"}]
            cue = RegelumSurvey(name, message, scale, fields)
            user_settings["JOBS_TESTING"] = cue.send()["cores"]["cores"]
    ok(
        "We're [bold magenta]done[/bold magenta]! Bear in mind that you can always change these settings by running regelum with [bold magenta]--configure[/bold magenta]."
    )
    OmegaConf.save(OmegaConf.create(user_settings), config_file)



"""Contains critics, which are integrated in scenarios (agents).

Remarks:

- All vectors are treated as of type [n,]
- All buffers are treated as of type [L, n] where each row is a vector
- Buffers are updated from bottom to top

"""

from unittest.mock import MagicMock
import numpy as np
from .utils import rg
from abc import ABC
from .optimizable import Optimizable
from .objective import temporal_difference_objective, temporal_difference_objective_full

try:
    import torch
except (ModuleNotFoundError, ImportError):
    torch = MagicMock()

from .model import Model, ModelNN
from typing import Optional, Union, List
from .optimizable import OptimizerConfig
from .data_buffers import DataBuffer
from .system import System, ComposedSystem
from regelum.typing import RgArray, Weights


class Critic(Optimizable, ABC):
    """Critic base class.

    A critic is an object that estimates or provides the value of a given action or state in a reinforcement learning problem.

    The critic estimates the value of an action by learning from past experience, typically through the optimization of a loss function.
    """

    def __init__(
        self,
        system: Union[System, ComposedSystem],
        model: Union[Model, ModelNN],
        td_n: int = 1,
        device: Union[str, torch.device] = "cpu",
        is_same_critic: bool = False,
        is_value_function: bool = False,
        is_on_policy: bool = False,
        optimizer_config: Optional[OptimizerConfig] = None,
        regularization_param: float = 0.0,
        action_bounds: Optional[Union[List, np.array]] = None,
        size_mesh: Optional[int] = None,
        discount_factor: float = 1.0,
        sampling_time: float = 0.01,
        is_full_iteration_epoch: bool = False,
    ):
        """Initialize a critic object.

        Args:
            system (Union[System, ComposedSystem]): system environment
                that is used in RL problem. The system is mainly used
                for extraction of `dim_observation` and `dim_action`
            model (Union[Model, ModelNN]): Model to use for the critic
            td_n (int, optional): How many running_objective terms to
                use for temporal difference, defaults to 1
            device (Union[str, torch.device], optional): Device to use
                for optimization, defaults to "cpu"
            is_same_critic (bool, optional): whether to use the
                undetached critic in temporal difference loss, defaults
                to False
            is_value_function (bool, optional): whether critis is Value
                function (i.e. depends on observation only) or
                Q-function (i.e. depends on observation and action). For
                `True` the critic is Value function. For `False` the
                critic is Q-function, defaults to False
            is_on_policy (bool, optional): whether critic is on policy
                or off policy optimized, defaults to False
            optimizer_config (Optional[OptimizerConfig], optional):
                optimizer configuration, defaults to None
            regularization_param (bool): L2 penalty for weights displacement
            when symbolic optimization engine is used.
            action_bounds (Optional[Union[List, np.array]], optional):
                action bounds. Needed for DQN algorithm for constraint
                optimization problem, defaults to None
            size_mesh (Optional[int], optional): action grid mesh size,
                needed for DQN algorithm for constraint optimization
                problem, defaults to None
            discount_factor (float, optional): discount factor to use in
                temporal difference loss, defaults to 1.0
            sampling_time (float, optional): scenario sampling time.
                Needed in temporal difference loos, defaults to 0.01
        """
        Optimizable.__init__(self, optimizer_config=optimizer_config)

        self.model = model
        self.discount_factor = discount_factor
        self.system = system
        self.sampling_time = sampling_time
        self.td_n = td_n
        self.device = device
        self.is_same_critic = is_same_critic
        self.is_value_function = is_value_function
        self.is_on_policy = is_on_policy
        self.action_bounds = (
            np.array(action_bounds) if action_bounds is not None else None
        )
        self.regularization_param = regularization_param
        self.size_mesh = size_mesh
        self.is_full_iteration_epoch = is_full_iteration_epoch

        self.initialize_optimize_procedure()

    def receive_estimated_state(self, state):
        self.state = state

    def __call__(
        self, *args, use_stored_weights: bool = False, weights: Optional[Weights] = None
    ) -> Union[RgArray, float]:
        """Compute the value of the critic function for a given observation and/or action.

        Args:
            *args (tuple): tuple of the form (observation, action) or
                (observation,)
            use_stored_weights (bool): flag indicating whether to use
                the stored weights of the critic model or the current
                weights

        Returns:
            Value of the critic function (either value or Q-function)
        """
        return self.model(*args, use_stored_weights=use_stored_weights, weights=weights)

    @property
    def weights(self):
        """Get the weights of the critic model."""
        return self.model.weights

    def update_weights(self, weights: Optional[Weights] = None) -> None:
        """Update the weights of the critic model.

        Args:
            weights: new weights to be used for the critic
                model, if not provided the method does nothing.
        """
        self.model.update_weights(weights)

    def cache_weights(self, weights: Optional[Weights] = None):
        """Store a copy of the current model weights.

        Args:
            weights: An optional ndarray of weights to store. If not
                provided, the current model weights are stored. Default
                is None.
        """
        self.model.cache_weights(weights)

    def restore_weights(self):
        """Restores the model weights to the cached weights."""
        self.model.restore_weights()

    def update_and_cache_weights(self, weights: Optional[Weights] = None):
        """Update the model's weights and cache the new values.

        Args:
            weights: new weights for the model (optional)
        """
        self.update_and_cache_weights(weights)

    def initialize_optimize_procedure(self):
        """Instantilize optimization procedure via Optimizable functionality."""
        self.batch_size = self.get_data_buffer_batch_size()
        (
            self.running_objective_var,
            self.critic_targets_var,
            self.critic_weights_var,
            self.critic_stored_weights_var,
        ) = (
            self.create_variable(
                self.batch_size,
                1,
                name="running_objective",
                is_constant=True,
            ),
            self.create_variable(
                self.batch_size,
                1,
                name="critic_targets",
                is_constant=True,
            ),
            self.create_variable(
                name="critic_weights",
                like=self.model.named_parameters,
            ),
            self.create_variable(
                name="critic_stored_weights",
                like=self.model.cache.weights,
                is_constant=True,
            ),
        )
        if not self.is_value_function:
            self.observation_action_var = self.create_variable(
                self.batch_size,
                self.system.dim_observation + self.system.dim_inputs,
                name="observation_action",
                is_constant=True,
            )
            self.critic_model_output = self.create_variable(
                self.batch_size,
                1,
                name="critic_model_output",
                is_nested_function=True,
                nested_variables=[
                    self.observation_action_var,
                    self.critic_stored_weights_var,
                    self.critic_weights_var,
                ],
            )
        else:
            self.observation_var = self.create_variable(
                self.batch_size,
                self.system.dim_observation,
                name="observation",
                is_constant=True,
            )
            self.critic_model_output = self.create_variable(
                self.batch_size,
                1,
                name="critic_model_output",
                is_nested_function=True,
                nested_variables=[
                    self.observation_var,
                    self.critic_stored_weights_var,
                    self.critic_weights_var,
                ],
            )

        if hasattr(self.model, "weight_bounds"):
            self.register_bounds(self.critic_weights_var, self.model.weight_bounds)

        if self.is_value_function:
            self.connect_source(
                connect_to=self.critic_model_output,
                func=self.model,
                source=self.observation_var,
                weights=self.critic_weights_var,
            )
            if (not self.is_same_critic) and self.is_on_policy:
                self.connect_source(
                    connect_to=self.critic_targets_var,
                    func=self.model.cache,
                    source=self.observation_var,
                    weights=self.critic_stored_weights_var,
                )
        else:
            self.connect_source(
                connect_to=self.critic_model_output,
                func=self.model,
                source=self.observation_action_var,
                weights=self.critic_weights_var,
            )
            if (not self.is_same_critic) and self.is_on_policy:
                self.connect_source(
                    connect_to=self.critic_targets_var,
                    func=self.model.cache,
                    source=self.observation_action_var,
                    weights=self.critic_stored_weights_var,
                )

        self.episode_id_var = self.create_variable(
            self.batch_size,
            1,
            name="episode_id",
            is_constant=True,
        )

        self.register_objective(
            func=self.objective_function,
            variables=[
                self.episode_id_var,
                self.running_objective_var,
                self.critic_model_output,
                self.critic_weights_var,
                self.critic_stored_weights_var,
            ]
            + (
                [self.critic_targets_var]
                if ((not self.is_same_critic) or (not self.is_on_policy))
                else []
            ),
        )

    def data_buffer_objective_keys(self) -> List[str]:
        """Return a list of `regelum.data_buffers.DataBuffer` keys to be used for the substitution to the objective function.

        Returns:
            List of keys.
        """
        if self.is_value_function:
            keys = ["observation", "running_objective", "episode_id"]
        else:
            keys = ["observation_action", "running_objective", "episode_id"]

        if not self.is_on_policy:
            keys.append("critic_targets")

        return keys

    def regularization_function(self, critic_stored_weights, critic_weights):
        return self.regularization_param * rg.sum(
            (critic_stored_weights - critic_weights) ** 2
        )

    def objective_function(
        self,
        critic_model_output,
        running_objective,
        critic_stored_weights,
        critic_weights,
        episode_id,
        critic_targets=None,
    ):
        if self.is_full_iteration_epoch:
            assert (
                self.kind == "tensor"
            ), "Full iteration epoch only supported for Torch critic"
            td_objective = temporal_difference_objective_full(
                critic_model_output=critic_model_output,
                running_objective=running_objective,
                episode_ids=episode_id,
                td_n=self.td_n,
                discount_factor=self.discount_factor,
                sampling_time=self.sampling_time,
                critic_targets=critic_targets,
            )
        else:
            td_objective = temporal_difference_objective(
                critic_model_output=critic_model_output,
                running_objective=running_objective,
                td_n=self.td_n,
                discount_factor=self.discount_factor,
                sampling_time=self.sampling_time,
                critic_targets=critic_targets,
            )
        if self.kind == "tensor":
            return td_objective

        else:
            return td_objective + self.regularization_function(
                critic_stored_weights=critic_stored_weights,
                critic_weights=critic_weights,
            )

    def update_data_buffer_with_optimal_policy_targets(self, data_buffer: DataBuffer):
        assert self.size_mesh is not None, "Specify size_mesh for off-policy critic"
        data_buffer_size = len(data_buffer)
        if data_buffer_size == 0:
            return
        observations = np.vstack(data_buffer.data["observation"])
        actions = np.random.uniform(
            self.action_bounds[:, 0],
            self.action_bounds[:, 1],
            size=(data_buffer_size, self.size_mesh, len(self.action_bounds)),
        )
        tiled_observations = np.vstack(
            [
                [np.tile(observation, (self.size_mesh, 1))]
                for observation in observations
            ]
        )
        model_inputs = torch.FloatTensor(
            np.concatenate((tiled_observations, actions), axis=2)
        ).to(self.device)

        values = self.model(model_inputs, use_stored_weights=True)
        data_buffer.update({"critic_targets": torch.min(values, axis=1).values.numpy()})

    def update_data_buffer_with_stored_weights_critic_output(
        self, data_buffer: DataBuffer
    ):
        self.model(data_buffer, use_stored_weights=True)

    def delete_critic_targets(self, data_buffer: DataBuffer):
        if "critic_targets" in data_buffer.keys():
            data_buffer.delete_key("critic_targets")

    def optimize(
        self, data_buffer, is_update_and_cache_weights=True, is_constrained=True
    ):
        if isinstance(self.model, ModelNN):
            self.model.to(self.device)
            self.model.cache.to(self.device)

        if not self.is_on_policy:
            self.update_data_buffer_with_optimal_policy_targets(data_buffer)

        opt_kwargs = data_buffer.get_optimization_kwargs(
            keys=self.data_buffer_objective_keys(),
            optimizer_config=self.optimizer_config,
        )
        weights = None
        if opt_kwargs is not None:
            if self.kind == "tensor":
                super().optimize(**opt_kwargs, is_constrained=is_constrained)
                if is_update_and_cache_weights:
                    self.model.update_and_cache_weights()
            elif self.kind == "symbolic":
                weights = super().optimize(
                    **opt_kwargs,
                    is_constrained=is_constrained,
                    critic_weights=self.model.weights,
                    critic_stored_weights=self.model.cache.weights,
                )["critic_weights"]
                if is_update_and_cache_weights:
                    self.model.update_and_cache_weights(weights)

        if not self.is_on_policy:
            self.delete_critic_targets(data_buffer)

        return weights


class CriticTrivial(Critic):
    """A mocked Critic object."""

    class TrivialModel:
        """A mocked Trivial model."""

        def named_parameters(self):
            return None

    def __init__(self, *args, **kwargs):
        """Instantiate a CriticTrivial object."""
        self.model = self.TrivialModel()

    def optimize(self, *args, **kwargs):
        pass

    def update_weights(self, *args, **kwargs):
        pass

    def cache_weights(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        pass

    @property
    def weights(self):
        return None


class CriticCALF(Critic):
    """Critic for CALF algorithm."""

    def __init__(
        self,
        system,
        model: Union[Model, ModelNN],
        is_same_critic: bool,
        is_value_function: bool,
        td_n: int = 1,
        predictor: Optional[Model] = None,
        optimizer_config: Optional[OptimizerConfig] = None,
        discount_factor: float = 1.0,
        sampling_time: float = 0.01,
        ######
        safe_decay_param: float = 1e-4,
        is_dynamic_decay_rate: bool = True,
        safe_policy=None,
        lb_parameter: float = 1e-6,
        ub_parameter: float = 1e3,
        regularization_param: float = 0,
    ):
        """Instantiate a CriticCALF object."""
        super().__init__(
            system=system,
            model=model,
            td_n=td_n,
            is_same_critic=is_same_critic,
            is_value_function=is_value_function,
            is_on_policy=True,
            optimizer_config=optimizer_config,
            discount_factor=discount_factor,
            sampling_time=sampling_time,
            action_bounds=None,
            regularization_param=regularization_param,
        )
        self.predictor = predictor
        self.safe_decay_param = safe_decay_param
        self.is_dynamic_decay_rate = is_dynamic_decay_rate
        if not self.is_dynamic_decay_rate:
            self.safe_decay_rate = self.safe_decay_param

        self.lb_parameter = lb_parameter
        self.ub_parameter = ub_parameter
        self.safe_policy = safe_policy

        self.observation_last_good_var = self.create_variable(
            self.batch_size,
            self.system.dim_observation,
            name="observation_last_good",
            is_constant=True,
        )
        self.prev_good_critic_var = self.create_variable(
            1,
            1,
            name="prev_good_critic",
            is_constant=True,
            is_nested_function=True,
            nested_variables=[
                self.observation_last_good_var,
                self.critic_stored_weights_var,
            ],
        )
        self.connect_source(
            connect_to=self.prev_good_critic_var,
            func=self.model.cache,
            source=self.observation_last_good_var,
            weights=self.critic_stored_weights_var,
        )

        self.register_constraint(
            self.CALF_decay_constraint_no_prediction,
            variables=[
                self.critic_model_output,
                self.prev_good_critic_var,
            ],
        )
        self.register_constraint(
            self.CALF_critic_lower_bound_constraint,
            variables=[self.critic_model_output, self.observation_var],
        )

        self.observation_last_good = None

    def data_buffer_objective_keys(self) -> List[str]:
        """Return a list of `regelum.data_buffers.DataBuffer` keys to be used for the substitution to the objective function.

        Returns:
            List of keys.
        """
        keys = super().data_buffer_objective_keys()
        keys.append("observation_last_good")
        return keys

    def CALF_decay_constraint_no_prediction(
        self,
        critic_model_output,
        prev_good_critic,
    ):
        stabilizing_constraint_violation = (
            critic_model_output[-1, :]
            - prev_good_critic[-2, :]
            + self.sampling_time * self.safe_decay_rate
        )
        return stabilizing_constraint_violation

    def CALF_critic_lower_bound_constraint(
        self, critic_model_output: RgArray, observation: RgArray
    ) -> RgArray:
        """Constraint that ensures that the value of the critic is above a certain lower bound.

        The lower bound is determined by the `current_observation` and a certain constant.

        Args:
            critic_model_output: output of a critic

        Returns:
            Constraint violation
        """
        self.lb_constraint_violation = (
            self.lb_parameter * rg.sum(observation[:-1, :] ** 2)
            - critic_model_output[-1, :]
        )
        return self.lb_constraint_violation



"""Module that contains general objectives functions that can be used by various entities of the framework.

For instance, a running objective can be used commonly by a generic optimal scenario, an actor, a critic, a logger, an animator, a scenario etc.

"""

from abc import ABC, abstractmethod

import regelum
from .model import Model, PerceptronWithTruncatedNormalNoise, ModelNN
from typing import Optional
import torch
from .utils import rg
from .predictor import Predictor
from regelum.typing import RgArray


class Objective(regelum.RegelumBase, ABC):
    """A base class for objective implementations."""

    def __init__(self):
        """Initialize an instance of Objective."""
        pass

    @abstractmethod
    def __call__(self):
        pass


class RunningObjective(Objective):
    """Running reward/cost.

    In minimzations problems, it is called cost or loss, say.
    """

    def __init__(self, model: Optional[Model] = None):
        """Initialize a RunningObjective instance.

        Args:
            model (function): function that calculates the running
                objective for a given observation and action.
        """
        self.model = (lambda observation, action: 0) if model is None else model

    def __call__(
        self, observation: RgArray, action: RgArray, is_save_batch_format: bool = False
    ) -> RgArray:
        """
        Calculate the running objective for a given observation and action, potentially formatting the output.

        This method computes the objective value based on the current state of the environment as represented
        by the observation and the action taken by an agent. It can also preprocess the output to be in batch
        format if specified.

        Args:
            observation: The current observation from the environment, which should reflect the
                current state. This is typically a numpy array or an array-like structure that contains
                numerical data representing the state of the environment the agent is interacting with.
            action: The action taken by the agent in response to the observation. Similar to the
                observation, this is typically a numpy array or an array-like structure containing numerical
                data that represents the action chosen by the agent.
            is_save_batch_format: A flag to determine if the output should be preprocessed
                into a batch format. When False, the method does not alter the output. When True, it will
                preprocess the output into a format that is suitable for batch processing, which is often
                required for machine learning models that process data in batches. Default is False.

        Returns:
            The computed running objective value, which is a numerical score representing the
                performance or the reward of the agent given the observation and the action taken.
                This value is usually used in reinforcement learning to guide the agent's learning process.
        """

        running_objective = self.model(observation, action)
        if not is_save_batch_format:
            return running_objective[0][0]

        return running_objective


class RunningObjectivePendulum(RunningObjective):
    def __call__(
        self, observation: RgArray, action: RgArray, is_save_batch_format: bool = False
    ) -> RgArray:
        model_res = super().__call__(observation, action, is_save_batch_format)
        return (
            model_res + 15 * observation[0, 0] ** 2
            if abs(observation[0, 0]) > 0.3
            else model_res + 15 * observation[0, 1] ** 2
        )


def reinforce_objective(
    policy_model: PerceptronWithTruncatedNormalNoise,
    observations: torch.FloatTensor,
    actions: torch.FloatTensor,
    tail_values: torch.FloatTensor,
    values: torch.FloatTensor,
    baselines: torch.FloatTensor,
    is_with_baseline: bool,
    is_do_not_let_the_past_distract_you: bool,
    N_episodes: int,
) -> torch.FloatTensor:
    """Calculate the surrogate objective for REINFORCE algorithm.

    Args:
        policy_model: The policy model that outputs log probabilities.
        observations: Observed states from the environment.
        actions: Actions taken in response to the observations.
        tail_values: The rewards-to-go or costs-to-go for each step in the episode.
        values: The total accumulated running objectives for each step in the episode.
        baselines: The baseline values used for variance reduction.
        is_with_baseline: If True, subtract baselines from the target objectives.
        is_do_not_let_the_past_distract_you: If True, use tail values instead of total values.
        N_episodes: The count of episodes over which the average is taken.

    Returns:
        The computed surrogate objective value as a float tensor.
    """
    log_pdfs = policy_model.log_pdf(observations, actions)
    if is_do_not_let_the_past_distract_you:
        target_objectives = tail_values
    else:
        target_objectives = values
    if is_with_baseline:
        target_objectives -= baselines

    return (log_pdfs * target_objectives).sum() / N_episodes


def get_gae_advantage(
    gae_lambda: float,
    running_objectives: torch.FloatTensor,
    values: torch.FloatTensor,
    times: torch.FloatTensor,
    discount_factor: float,
    sampling_time: float,
) -> torch.FloatTensor:
    """Calculate the Generalized Advantage Estimation (GAE) advantage.

    Args:
        gae_lambda: The GAE lambda parameter.
        running_objectives: The running objectives tensor.
        values: The critic values tensor.
        times: The timestamps tensor.
        discount_factor: The discount factor.
        sampling_time: The sampling time between actions

    Returns:
        GAE advantage tensor.
    """
    deltas = (
        running_objectives[:-1]
        + discount_factor**sampling_time * values[1:]
        - values[:-1]
    )
    if gae_lambda == 0.0:
        advantages = deltas
    else:
        gae_discount_factors = (gae_lambda * discount_factor) ** times[:-1]
        reversed_gae_discounted_deltas = torch.flip(
            gae_discount_factors * deltas, dims=[0, 1]
        )
        advantages = (
            torch.flip(reversed_gae_discounted_deltas.cumsum(dim=0), dims=[0, 1])
            / gae_discount_factors
        )
    return advantages


def sdpg_objective(
    policy_model: PerceptronWithTruncatedNormalNoise,
    critic_model: ModelNN,
    observations: torch.FloatTensor,
    actions: torch.FloatTensor,
    times: torch.FloatTensor,
    episode_ids: torch.LongTensor,
    discount_factor: float,
    N_episodes: int,
    running_objectives: torch.FloatTensor,
    sampling_time: float,
    is_normalize_advantages: bool,
    gae_lambda: float,
) -> torch.FloatTensor:
    """Calculate the surrogate objective for the Stochastic Deterministic Policy Gradient (SDPG) algorithm.

    Args:
        policy_model: The policy model that outputs log probabilities.
        critic_model: The critic model that estimates Q-values for observation-action pairs.
        observations: A tensor of observed states from the environment.
        actions: A tensor of actions taken by the agent.
        times: A tensor of time steps corresponding to each observation-action pair.
        episode_ids: A tensor of episode identifiers.
        discount_factor: A scalar discount factor for future rewards, typically between 0 and 1.
        N_episodes: The number of episodes to average over for the objective calculation.
        running_objectives: A tensor of running objective values up to the current time step.
        sampling_time: The time step used for sampling in the environment.
        is_normalize_advantages: If True, normalize the advantage estimates.
        gae_lambda: The lambda parameter for GAE, controlling the trade-off between bias and variance.


    Returns:
        SDPG surrogate objective.
    """
    critic_values = critic_model(observations)
    log_pdfs = policy_model.log_pdf(observations, actions).reshape(-1)

    objective = 0.0
    for episode_idx in torch.unique(episode_ids):
        mask = (episode_ids == episode_idx).reshape(-1)
        advantages = get_gae_advantage(
            gae_lambda=gae_lambda,
            running_objectives=running_objectives[mask],
            values=critic_values[mask],
            times=times[mask],
            discount_factor=discount_factor,
            sampling_time=sampling_time,
        ).reshape(-1)

        if is_normalize_advantages:
            advantages = (advantages - advantages.mean()) / (advantages.std() + 1e-8)

        objective += (
            discount_factor ** times[mask][:-1].reshape(-1)
            * advantages
            * log_pdfs[mask][:-1]
        ).sum()

    return objective / N_episodes


def ppo_objective(
    policy_model: PerceptronWithTruncatedNormalNoise,
    critic_model: ModelNN,
    observations: torch.FloatTensor,
    actions: torch.FloatTensor,
    times: torch.FloatTensor,
    episode_ids: torch.LongTensor,
    discount_factor: float,
    N_episodes: int,
    running_objectives: torch.FloatTensor,
    cliprange: float,
    initial_log_probs: torch.FloatTensor,
    running_objective_type: str,
    sampling_time: float,
    gae_lambda: float,
    is_normalize_advantages: bool = True,
    entropy_coeff: float = 0.0,
) -> torch.FloatTensor:
    """Calculate PPO objective.

    Args:
        policy_model: The neural network model representing the policy.
        critic_model: The neural network model representing the value function (critic).
        observations: A tensor of observations from the environment.
        actions: A tensor of actions taken by the agent.
        times: A tensor with timestamps for each observation-action pair.
        episode_ids: A tensor with unique identifiers for each episode.
        discount_factor: The factor by which future rewards are discounted.
        N_episodes: The total number of episodes over which the objective is averaged.
        running_objectives: A tensor of accumulated rewards or costs for each timestep.
        cliprange: The range for clipping the probability ratio in the objective function.
        initial_log_probs: The log probabilities of taking the actions at the time
            of sampling, under the policy model before the update.
        running_objective_type (str): Indicates whether the running objectives are 'cost' or 'reward'.
        sampling_time: The timestep used for sampling in the environment.
        gae_lambda: The lambda parameter for GAE, controlling the trade-off between bias and variance.
        is_normalize_advantages: Flag indicating whether to normalize advantage estimates.

    Returns:
        objective for PPO
    """
    assert (
        running_objective_type == "cost" or running_objective_type == "reward"
    ), "running_objective_type can be either 'cost' or 'reward'"

    critic_values = critic_model(observations)
    prob_ratios = torch.exp(
        policy_model.log_pdf(observations, actions) - initial_log_probs.reshape(-1)
    ).reshape(-1, 1)
    if hasattr(policy_model, "entropy"):
        entropies = entropy_coeff * policy_model.entropy(observations).reshape(-1, 1)
    else:
        entropies = torch.zeros_like(prob_ratios)
    clipped_prob_ratios = torch.clamp(prob_ratios, 1 - cliprange, 1 + cliprange)
    objective_value = 0.0
    for episode_idx in torch.unique(episode_ids):
        mask = episode_ids.reshape(-1) == episode_idx
        advantages = get_gae_advantage(
            gae_lambda=gae_lambda,
            running_objectives=running_objectives[mask],
            values=critic_values[mask],
            times=times[mask],
            discount_factor=discount_factor,
            sampling_time=sampling_time,
        )
        if is_normalize_advantages:
            advantages = (advantages - advantages.mean()) / (advantages.std() + 1e-8)

        objective_value += (
            torch.sum(
                (discount_factor ** times[mask][:-1])
                * (
                    torch.maximum(
                        advantages * prob_ratios[mask][:-1],
                        advantages * clipped_prob_ratios[mask][:-1],
                    )
                    - entropies[mask][:-1]
                    if running_objective_type == "cost"
                    else torch.minimum(
                        advantages * prob_ratios[mask][:-1],
                        advantages * clipped_prob_ratios[mask][:-1],
                    )
                    + entropies[mask][:-1]
                )
            )
            / N_episodes
        )

    return objective_value


def ddpg_objective(
    policy_model: ModelNN,
    critic_model: ModelNN,
    observations: torch.FloatTensor,
) -> torch.FloatTensor:
    """Calculate the objective value for the DDPG algorithm.

    Args:
        policy_model: The policy model that generates actions based on observations.
        critic_model: The critic model that approximates the Q-function.
        observations: The batch of observations.

    Returns:
        torch.FloatTensor: The objective value.
    """
    return critic_model(
        observations, policy_model.forward(observations, is_means_only=True)
    ).mean()


def temporal_difference_objective_full(
    critic_model_output: ModelNN,
    running_objective: torch.FloatTensor,
    td_n: int,
    discount_factor: float,
    sampling_time: float,
    episode_ids: torch.LongTensor,
    critic_targets: Optional[torch.FloatTensor] = None,
) -> torch.FloatTensor:
    objective = 0.0
    n_iterations = torch.unique(episode_ids).shape[0]
    for episode_id in torch.unique(episode_ids):
        mask = (episode_ids == episode_id).reshape(-1)
        objective += (
            temporal_difference_objective(
                critic_model_output=critic_model_output[mask],
                running_objective=running_objective[mask],
                td_n=td_n,
                discount_factor=discount_factor,
                sampling_time=sampling_time,
                critic_targets=(
                    critic_targets[mask] if critic_targets is not None else None
                ),
            )
            / n_iterations
        )
    return objective


def temporal_difference_objective(
    critic_model_output: ModelNN,
    running_objective: torch.FloatTensor,
    td_n: int,
    discount_factor: float,
    sampling_time: float,
    critic_targets: Optional[torch.FloatTensor] = None,
) -> torch.FloatTensor:
    """Calculate temporal difference objective.

    Args:
        critic_model (ModelNN): Q function model.
        observation (torch.FloatTensor): Batch of observations.
        action (torch.FloatTensor): Batch of actions.
        running_objective (torch.FloatTensor): Batch of running
            objectives.
        td_n (int): Number of temporal difference steps.
        discount_factor (float): Discount factor for future running
            objectives.
        device (Union[str, torch.device]): Device to proceed
            computations on.
        sampling_time (float): Sampling time for discounting.
        is_use_same_critic (bool): Whether to use the critic model from
            the previous iteration or not (`True` for not using, `False`
            for using).

    Returns:
        torch.FloatTensor: objective value
    """
    batch_size = running_objective.shape[0]
    assert batch_size > td_n, f"batch size {batch_size} too small for such td_n {td_n}"
    discount_factors = rg.array(
        [[discount_factor ** (sampling_time * i)] for i in range(td_n)],
        prototype=running_objective,
        _force_numeric=True,
    )
    discounted_tdn_sum_of_running_objectives = rg.vstack(
        [
            rg.sum(running_objective[i : i + td_n, :] * discount_factors)
            for i in range(batch_size - td_n)
        ]
    )

    if critic_targets is None:
        critic_targets = critic_model_output[td_n:, :]
    else:
        critic_targets = critic_targets[td_n:, :]

    temporal_difference = rg.mean(
        (
            critic_model_output[:-td_n, :]
            - discounted_tdn_sum_of_running_objectives
            - discount_factor ** (td_n * sampling_time) * critic_targets
        )
        ** 2
    )
    return temporal_difference


def mpc_objective(
    observation,
    estimated_state,
    policy_model_weights,
    predictor: Predictor,
    running_objective,
    model,
    prediction_horizon,
    discount_factor=1.0,
):
    (
        state_sequence_predicted,
        action_sequence_predicted,
    ) = predictor.predict_state_sequence_from_model(
        estimated_state,
        prediction_horizon=prediction_horizon,
        model=model,
        model_weights=policy_model_weights,
        is_predict_last=False,
    )
    observation_sequence_predicted = predictor.system.get_observation(
        None,
        state=state_sequence_predicted,
        inputs=action_sequence_predicted,
        is_batch=True,
    )
    observation_sequence = rg.vstack(
        (rg.force_row(observation), observation_sequence_predicted)
    )

    running_objectives = running_objective(
        observation_sequence, action_sequence_predicted, is_save_batch_format=True
    )

    if discount_factor < 1.0:
        discount_factors = rg.array(
            [
                [discount_factor ** (predictor.pred_step_size * i)]
                for i in range(prediction_horizon + 1)
            ],
            prototype=observation,
            _force_numeric=True,
        )
        mpc_objective_value = rg.sum(running_objectives * discount_factors)
    else:
        mpc_objective_value = rg.sum(running_objectives)

    return mpc_objective_value


def rpv_objective(
    observation,
    estimated_state,
    policy_model_weights,
    predictor: Predictor,
    running_objective,
    model,
    prediction_horizon,
    discount_factor,
    critic,
    critic_weights,
):
    rpv_objective_value = 0
    (
        state_sequence_predicted,
        action_sequence_predicted,
    ) = predictor.predict_state_sequence_from_model(
        estimated_state,
        prediction_horizon=prediction_horizon,
        model=model,
        model_weights=policy_model_weights,
        is_predict_last=True,
    )
    observation_sequence_predicted = predictor.system.get_observation(
        None,
        state=state_sequence_predicted,
        inputs=action_sequence_predicted,
        is_batch=True,
    )
    observation_sequence = rg.vstack(
        (rg.force_row(observation), observation_sequence_predicted[:-1, :])
    )

    discount_factors = rg.array(
        [
            [discount_factor ** (predictor.pred_step_size * i)]
            for i in range(prediction_horizon)
        ],
        prototype=observation,
        _force_numeric=True,
    )

    rpv_objective_value = rg.sum(
        discount_factors
        * running_objective(
            observation_sequence, action_sequence_predicted, is_save_batch_format=True
        )
    )

    observation_last = observation_sequence_predicted[-1, :]
    rpv_objective_value += rg.sum(
        discount_factor ** (predictor.pred_step_size * prediction_horizon)
        * critic(observation_last, weights=critic_weights)
    )

    return rpv_objective_value


def rql_objective(
    observation,
    estimated_state,
    policy_model_weights,
    predictor: Predictor,
    running_objective,
    model,
    prediction_horizon,
    discount_factor,
    critic,
    critic_weights,
):
    rql_objective_value = 0
    (
        state_sequence_predicted,
        action_sequence_predicted,
    ) = predictor.predict_state_sequence_from_model(
        estimated_state,
        prediction_horizon=prediction_horizon,
        model=model,
        model_weights=policy_model_weights,
        is_predict_last=False,
    )

    if prediction_horizon == 0:
        return critic(observation, action_sequence_predicted, weights=critic_weights)
    observation_sequence_predicted = predictor.system.get_observation(
        None,
        state=state_sequence_predicted,
        inputs=action_sequence_predicted,
        is_batch=True,
    )
    observation_sequence = rg.vstack(
        (rg.force_row(observation), observation_sequence_predicted[:-1, :])
    )
    discount_factors = rg.array(
        [
            [discount_factor ** (predictor.pred_step_size * i)]
            for i in range(prediction_horizon)
        ],
        prototype=observation,
        _force_numeric=True,
    )
    rql_objective_value = rg.sum(
        discount_factors
        * running_objective(
            observation_sequence,
            action_sequence_predicted[:-1, :],
            is_save_batch_format=True,
        )
    )

    observation_last = state_sequence_predicted[-1, :]
    rql_objective_value += rg.sum(
        discount_factor ** (predictor.pred_step_size * prediction_horizon)
        * critic(
            observation_last, action_sequence_predicted[-1, :], weights=critic_weights
        )
    )

    return rql_objective_value


def sql_objective(
    observation,
    estimated_state,
    policy_model_weights,
    predictor: Predictor,
    model,
    prediction_horizon,
    critic,
    critic_weights=None,
):
    (
        state_sequence_predicted,
        action_sequence_predicted,
    ) = predictor.predict_state_sequence_from_model(
        estimated_state,
        prediction_horizon=prediction_horizon,
        model=model,
        model_weights=policy_model_weights,
        is_predict_last=False,
    )
    observation_sequence_predicted = predictor.system.get_observation(
        None,
        state=state_sequence_predicted,
        inputs=action_sequence_predicted,
        is_batch=True,
    )
    observation_sequence = rg.vstack(
        (rg.force_row(observation), observation_sequence_predicted)
    )

    sql_objective_value = rg.sum(
        rg.sum(
            critic(
                observation_sequence, action_sequence_predicted, weights=critic_weights
            )
        )
    )

    return sql_objective_value



"""Module contains a set of standard hooks for optimizable objects."""

import regelum
from typing import Callable, List
from types import GeneratorType


def get_data_hook(var):
    def var_hook(whatever):
        return var.data

    return var_hook


def requires_grad(variable_data):
    if isinstance(variable_data, (List, GeneratorType)):
        for datum in variable_data:
            if isinstance(datum, tuple):
                datum[1].requires_grad_(True)
            else:
                datum.requires_grad_(True)
    elif isinstance(variable_data, Callable):
        return requires_grad(variable_data())
    else:
        variable_data.requires_grad_(True)
    return variable_data


def detach(variable_data):
    if isinstance(variable_data, (List, GeneratorType)):
        for datum in variable_data:
            if isinstance(datum, tuple):
                datum[1].requires_grad_(False)
            else:
                datum.requires_grad_(False)
    elif isinstance(variable_data, Callable):
        return detach(variable_data())
    else:
        variable_data.requires_grad_(False)
    return variable_data


def mutate_metadata(self, new_metadata, tag="default"):
    def metadata_mutator(whatever):
        return new_metadata

    hook = regelum.Hook(metadata_mutator, metadata=tag, act_on="metadata")
    return hook


def data_closure(gen_method):
    def hook_data(whatever):
        if callable(gen_method):
            gen_tmp = gen_method()
            for x in gen_tmp:
                if x[1].requires_grad:
                    yield x[1]
        else:
            yield gen_method

    return hook_data


def metadata_closure(gen_method):
    def hook_metadata(whatever):
        gen_tmp = gen_method()
        for x in gen_tmp:
            yield x

    return hook_metadata



"""Contains a generic interface for systems (environments) as well as concrete systems as realizations of the former.

Note:
    The classes provided here do not perform any simulation mechanics itself.
    It contains only the update rule as defined in this section. The actual simulation routine is executed by a
    separate [`Simulator`][regelum.simulator] class, which leverages the update rule to create the corresponding integration scheme.
"""

from __future__ import annotations
import numpy as np
import casadi as cs
import torch
import regelum
from abc import ABC, abstractmethod

from . import callback
# from . import animation
from .utils import rg
from typing import Any, Optional, Union, List, Dict, Tuple
from typing_extensions import Self
from regelum.typing import RgArray


class SystemInterface(regelum.RegelumBase, ABC):
    """Abstract base class defining the interface for system environments.

    This interface sets the foundation for specific system implementations, providing properties for common attributes
    and abstract methods that must be implemented by subclasses to define system dynamics and observations.

    Attributes:
        _name: Name identifier for the system.
        _system_type: A string representing the type of the system (e.g., `"diff_eqn"`).
        _dim_state: The number of state variables of the system.
        _dim_inputs: The number of input variables, typically control inputs. In regelum we commonly use inputs parameter for actions.
        _dim_observation: The number of observation variables.
        _parameters: A dictionary of system parameters.
        _observation_naming: A list of strings naming each observation dimension.
        _states_naming: A list of strings naming each state dimension.
        _inputs_naming: A list of strings naming each input dimension.
        _action_bounds: A list of pairs defining the lower and upper bounds for each action dimension.
            Action bounds are defined as a list of [min, max] pairs for each action dimension
    """

    _name: Optional[str] = None
    _system_type: Optional[str] = None
    _dim_state: Optional[int] = None
    _dim_inputs: Optional[int] = None
    _dim_observation: Optional[int] = None
    _parameters: Optional[Dict[str, float]] = {}
    _observation_naming: Optional[List[str]] = None
    _state_naming: Optional[List[str]] = None
    _inputs_naming: Optional[List[str]] = None
    _action_bounds: Optional[List[List[float]]] = None

    @property
    def name(self) -> str:
        if self._name is None:
            raise ValueError("system._name should be set")
        return self._name

    @property
    def system_type(self) -> str:
        if self._system_type is None:
            raise ValueError("system._system_type should be set")
        return self._system_type

    @property
    def dim_state(self) -> int:
        if self._dim_state is None:
            raise ValueError("system._dim_state should be set")
        return self._dim_state

    @property
    def dim_observation(self) -> int:
        if self._dim_observation is None:
            raise ValueError("system._dim_observation should be set")
        return self._dim_observation

    @property
    def dim_inputs(self) -> int:
        if self._dim_inputs is None:
            raise ValueError("system._dim_inputs should be set")
        return self._dim_inputs

    @property
    def parameters(self) -> dict:
        return self._parameters

    @property
    def state_naming(self):
        return self._state_naming

    @property
    def observation_naming(self):
        return self._observation_naming

    @property
    def inputs_naming(self):
        return self._inputs_naming

    @property
    def action_bounds(self):
        return self._action_bounds

    @abstractmethod
    def _compute_state_dynamics(
        self, time: Union[float, cs.MX], state: RgArray, inputs: RgArray
    ) -> RgArray:
        """Compute the derivative of the state with respect to time given the current state and inputs.

        This abstract method must be implemented by the inheriting system classes to define the specific system
        dynamics for the simulation. It provides the fundamental mathematical model that describes how the state of the
        system evolves over time in response to (e.g control or noisy) inputs.

        It is intended to be used internally by the [`SystemInterface.compute_state_dynamics`][regelum.system.SystemInterface.compute_state_dynamics] method
        which provides a public interface, handling dimensionality enforcement and batch processing if necessary. The
        `_compute_state_dynamics` method ensures that the core dynamics calculations are encapsulated within each
        system's subclass, promoting a clear separation of the dynamics computation from any input preprocessing or
        other wrapping functionality.

        The implementation of this method in derived classes will typically involve evaluating the differential
        equations that govern the system's dynamics, returning the rate of change of the system's state variables.

        Args:
            time: The current simulation time.
            state: The current state vector of the system, where state[i] corresponds to the i-th state variable.
            inputs: The current input vector to the system, where inputs[i] corresponds to the i-th control input.

        Returns:
            The computed rate of change of the system's state vector.

        Note:
            This method does not update the state of the system. It only returns the derivative which may then be used
            by a numerical integrator to simulate the system's behavior over time.
        """
        pass

    def compute_state_dynamics(
        self,
        time: Union[float, cs.MX],
        state: RgArray,
        inputs: RgArray,
        _native_dim: bool = False,
    ) -> RgArray:
        """Computes the state dynamics of the system given the current state and inputs.

        This method is a public wrapper for the abstract method `_compute_state_dynamics`, which must be implemented by
        derived classes.

        Arguments:
            time: Current simulation time.
            state: The current state of the system.
            inputs: The current inputs to the system.
            _native_dim (optional): A flag indicating whether the method should maintain the original dimensionality of its inputs.

        Returns:
            The computed state dynamics.
        """
        if not _native_dim:
            return rg.force_row(
                self._compute_state_dynamics(
                    time, rg.force_column(state), rg.force_column(inputs)
                )
            )
        else:
            return self._compute_state_dynamics(time, state, inputs)

    @abstractmethod
    def _get_observation(
        self, time: Union[float, cs.MX], state: RgArray, inputs: RgArray
    ) -> RgArray:
        """Generate the observation of the system given the current state and inputs.

        This abstract method must be implemented by subclasses and is central to the observation-generating process of
        the system. In reinforcement learning and control systems, observations are the perceivable parts of the
        system's state and inputs that are accessible at a particular point in time. These observations are used by an
        agent or a controller to make decisions or by evaluation tools to assess the system's performance.

        The `_get_observation` method specifically handles the transformation of the system's state and inputs into a
        format best suited for decision-making or evaluation. This transformation may include, but is not limited to,
        sensor models, noise injection, feature extraction, or scaling.

        Args:
            time: The current simulation time or step.
            state: The current state vector of the system.
            inputs: The current input vector given to the system.

        Returns:
            An observation vector representing the perceived state of the system at the current time.

        Note:
            Implementing classes should define what makes up an observation in the context of their specific system.
            The method should aim to represent realistically perceivable quantities
        """
        pass

    def get_observation(
        self,
        time: Union[float, cs.MX],
        state: RgArray,
        inputs: RgArray,
        _native_dim: bool = False,
        is_batch: bool = False,
    ) -> RgArray:
        """Public method that wraps around the `_get_observation` abstract method to produce observations.

        This public interface method calls the internal abstract method
        [`_get_observation`][regelum.system.SystemInterface._get_observation] and applies additional
        formatting or dimensionality adjustments. It provides flexibility in handling observations on a per-instance or
        batch processing basis, which can be essential for efficiency in various simulations and algorithm
        implementations.

        Args:
            time: The current simulation time or step.
            state: The current state of the system.
            inputs: The current inputs given to the system.
            _native_dim: A flag indicating if the native dimensionality from `_get_observation` should be preserved.
                Defaults to False, which reshapes the output to align with expected dimensions in batch processing or
                other uses.
            is_batch: A flag indicating if the method should handle inputs and states in a batched form, processing multiple instances simultaneously.
                This is useful for vectorized operations that consider multiple simulation instances at once.

        Returns:
            Formatted observation vector(s) consistent with the system's defined observation space structure and
            sizing.
        """
        if not is_batch:
            if not _native_dim:
                return rg.force_row(
                    self._get_observation(
                        time, rg.force_column(state), rg.force_column(inputs)
                    )
                )
            else:
                return self._get_observation(time, state, inputs)
        else:
            observations = rg.zeros(
                (state.shape[0], self.dim_observation), prototype=state
            )
            for i in range(state.shape[0]):
                observations[i, :] = self.get_observation(
                    time=time,
                    state=state[i, :],
                    inputs=inputs[i, :],
                    is_batch=False,
                )
            return observations

    def apply_action_bounds(self, action: np.array) -> np.array:
        """Apply the system's predefined action bounds to the given action vector.

        This method is responsible for ensuring that all actions executed by the system are within the specified safe
        or valid operating ranges. The enforcement of action bounds is a critical safety feature for many physical
        systems and a way to align virtual agent behavior with feasible actions in reinforcement learning environments.

        Args:
            action: A vector of actions proposed for the system, which may include elements outside the permissible bounds.

        Returns:
            A clipped action vector where each element is within the predefined action bounds of the system. If no
            bounds are defined, the input vector is returned unchanged.

        Note:
            The action bounds are typically set during the instantiation of a system class and represent the maximum
            and minimum values allowed for each action variable.
        """
        return (
            np.clip(
                action,
                np.array(self.action_bounds)[:, 0],
                np.array(self.action_bounds)[:, 1],
            )
            if self.action_bounds is not None
            else action
        )


class ComposedSystem(SystemInterface):
    """Represents a composed system created from combining two subsystems.

    The composed system allows for creating complex system dynamics by combining simpler subsystems. The outputs of one
    subsystem can be connected to the inputs of another through an input/output mapping.

    Attributes are inherited and further defined based on the combination of subsystems.

    An instance of this class is being created automatically when applying a `@` operation on two systems.

    The constructor creates a complex system by interconnecting the outputs of one system (`sys_left`) to the inputs of
    another (`sys_right`) based on an optional input/output mapping. The combination can be thought of as a way of
    layering or cascading systems to create a more intricate overall system architecture.
    """

    def __call__(self) -> Any:
        """Needed for instantiation through _target_ in hydra configs

        Returns:
            Self
        """
        return self

    def __init__(
        self,
        sys_left: Union[System, Self],
        sys_right: Union[System, Self],
        io_mapping: Optional[List[int]] = None,
        output_mode: str = "right",
        state_naming: List[str] = None,
        inputs_naming: List[str] = None,
        observation_naming: List[str] = None,
        action_bounds: List[List[float]] = None,
    ):
        """Initialize a ComposedSystem by combining two subsystems.

        Args:
            sys_left: The first subsystem whose outputs can be connected to another system's inputs.
            sys_right: The second subsystem which can receive connections from the 'left' system.
            io_mapping: An optional list defining the connections between the 'left'
                system state components and the 'right' system input components. Each item represents an index from the
                left system whose output is to be fed as an input to the same index in the right system. Defaults to
                None, which implies a direct connection of all available states to inputs.
            output_mode: A string specifying how to aggregate and output the observations from the composed system.
                Options are `"state"`, `"right"`, or `"both"`. Defaults to "right" which returns the observation of the right
                system only.
            state_naming: Custom naming for the state variables of the ComposedSystem.
            inputs_naming: Custom naming for the input variables of the ComposedSystem.
            observation_naming: Custom naming for the observation variables of the ComposedSystem.
            action_bounds: Action bounds for the composite system,
                defined as a list of [min, max] pairs for each action dimension.

        Note:
            The `output_mode` parameter affects how the state and observation dimensions are defined in the
            ComposedSystem. The composition modifies the routing of signals between systems and, depending on the mode,
            changes the reported system observations.
        """
        self._state_naming = state_naming
        self._inputs_naming = inputs_naming
        self._observation_naming = observation_naming
        self._action_bounds = (
            sys_left.action_bounds if action_bounds is None else action_bounds
        )

        if io_mapping is None:
            io_mapping = np.arange(min(sys_left.dim_state, sys_right.dim_inputs))

        assert output_mode in [
            "state",
            "right",
            "both",
        ], "output_mode must be 'state', 'right' or 'both'"

        if "diff_eqn" in [sys_left.system_type, sys_right.system_type]:
            self._system_type = "diff_eqn"
        else:
            self._system_type = sys_right.system_type

        self.sys_left = sys_left
        self.sys_right = sys_right
        self._parameters = sys_left.parameters | sys_right.parameters
        self._dim_state = self.sys_right.dim_state + self.sys_left.dim_state
        if output_mode == "state":
            self._dim_observation = self.sys_left.dim_state + self.sys_right.dim_state
        elif output_mode == "right":
            self._dim_observation = self.sys_right.dim_observation
        elif output_mode == "both":
            self._dim_observation = (
                self.sys_left.dim_observation + self.sys_right.dim_observation
            )
        self.rout_idx, self.occupied_idx = self.__get_routing(io_mapping)
        self._dim_inputs = (
            self.sys_right.dim_inputs
            + self.sys_left.dim_inputs
            - len(self.occupied_idx)
        )
        self._name = self.sys_left.name + " + " + self.sys_right.name

        self.free_right_input_indices = np.setdiff1d(
            np.arange(self.sys_right.dim_inputs).astype(int),
            self.occupied_idx.astype(int),
        )
        self.output_mode = output_mode
        self.forward_permutation = np.arange(self.dim_observation).astype(int)
        self.inverse_permutation = np.arange(self.dim_observation).astype(int)

    @staticmethod
    def __get_routing(io_mapping: List[int]) -> Tuple[np.array, np.array]:
        """Generate routing indices for state-to-input mapping of connected subsystems.

        This internal static method processes the input/output mapping provided during the composition of two subsystems.
        The method converts the mapping into index arrays that can be used to route the outputs of the left subsystem
        to the inputs of the right subsystem during the computation of state dynamics and observations.

        Args:
            io_mapping: A list defining how outputs (state) of the left subsystem are to be routed
                to inputs of the right subsystem. Each list item represents an index from the
                left system's output matched to an index in the right system's input.

        Returns:
            Two numpy arrays, where the first one contains indices of the left system's
                outputs, and the second one contains indices of the right system's inputs.
        """
        io_mapping_extended = []
        if len(io_mapping) > 0:
            for i, outputs in enumerate(io_mapping):
                assert type(outputs) in [
                    tuple,
                    int,
                    np.int32,
                    np.int64,
                    None,
                ], (
                    "io_mapping must be a list of ints or tuples or Nones. "
                    + f"However a value of type {type(outputs)} was provided."
                )
                if not isinstance(outputs, tuple):
                    outputs = (outputs,)
                elif outputs is None:
                    continue
                for output in outputs:
                    io_mapping_extended.append([i, output])

            io_mapping_extended = sorted(io_mapping_extended, key=lambda x: x[1])
            rout_idx, occupied_idx = rg.array(io_mapping_extended).astype(int).T
        else:
            rout_idx, occupied_idx = rg.array([]), rg.array([])
        return rout_idx, occupied_idx

    def _compute_state_dynamics(
        self,
        time: Union[float, cs.MX],
        state: RgArray,
        inputs: RgArray,
        _native_dim: bool = True,
    ) -> RgArray:
        """Compute the state dynamics of the composed system based on connected subsystems' dynamics.

        This method combines the individual state dynamics computations of the left and right subsystems that form the composed system.
        It appropriately routes outputs from the left subsystem as inputs to the right subsystem based on the established input/output mapping.
        State dynamics of each subsystem are then calculated individually, and the results are consolidated into a single state dynamics vector
        representing the dynamics of the entire composed system.

        Args:
            time: The current simulation time.
            state: The combined state vector of both subsystems, concatenated vertically (left subsystem states followed by right subsystem states).
            inputs: The input vector provided to the composed system.
            _native_dim: A boolean indicating whether to preserve the native dimensionality of the state and input vectors.

        Returns:
            A state dynamics vector for the whole composed system reflecting the rates of change of state variables.
        """
        state = rg.array(state, prototype=state)
        inputs = rg.array(inputs, prototype=state)
        state = state[self.forward_permutation]

        state_for_left, state_for_right = (
            state[: self.sys_left.dim_state],
            state[self.sys_left.dim_state :],
        )
        inputs_for_left = inputs[: self.sys_left.dim_inputs]
        dstate_of_left = rg.squeeze(
            self.sys_left.compute_state_dynamics(
                time=time,
                state=state_for_left,
                inputs=inputs_for_left,
                _native_dim=_native_dim,
            )
        )
        outputs_of_left = rg.squeeze(
            self.sys_left.get_observation(
                time=time,
                state=state_for_left,
                inputs=inputs_for_left,
                _native_dim=_native_dim,
            )
        )

        inputs_for_right = rg.zeros(
            self.sys_right.dim_inputs,
            prototype=(state, inputs),
        )
        if len(self.occupied_idx) > 0:
            inputs_for_right[self.occupied_idx] = outputs_of_left[self.rout_idx]
        inputs_for_right[self.free_right_input_indices] = rg.reshape(
            inputs[self.sys_left.dim_inputs :],
            rg.shape(inputs_for_right[self.free_right_input_indices]),
        )

        dstate_of_right = rg.squeeze(
            self.sys_right.compute_state_dynamics(
                time=time,
                state=state_for_right,
                inputs=inputs_for_right,
                _native_dim=_native_dim,
            )
        )
        final_dstate_vector = (
            rg.hstack((rg.force_row(dstate_of_left), rg.force_row(dstate_of_right)))
            if not isinstance(dstate_of_left, (np.ndarray, torch.Tensor))
            else rg.hstack((dstate_of_left, dstate_of_right))
        )

        assert (
            final_dstate_vector is not None
        ), f"final dstate_vector of system {self.name} is None"
        final_dstate_vector = final_dstate_vector[self.inverse_permutation]
        if not _native_dim:
            final_dstate_vector = rg.force_row(final_dstate_vector)
        return final_dstate_vector

    def _get_observation(
        self,
        time: Union[float, cs.MX],
        state: RgArray,
        inputs: RgArray,
        _native_dim: bool = False,
    ) -> RgArray:
        """Compute the observation of the composed system based on the observations from connected subsystems.

        This method uses the internal observations generation of the left and right subsystems to produce the overall composed system's observation.
        Depending on the chosen output mode ('state', 'right', or 'both'), the method prepares the observation vector by either using the state of the
        composed system, the observation from the right subsystem only, or the concatenation of both subsystems' observations.

        Args:
            time: The current simulation time.
            state: The combined state vector for both subsystems.
            inputs: The input vector provided to the composed system.
            _native_dim: A boolean flag to indicate if the method should maintain the native dimensionality of its inputs.

        Returns:
            An observation vector as per the defined output mode of the composed system.
        """
        state = rg.array(state, prototype=state)
        inputs = rg.array(inputs, prototype=state)

        inputs_for_left = inputs[: self.sys_left.dim_inputs]
        state_for_left, state_for_right = (
            state[: self.sys_left.dim_state],
            state[self.sys_left.dim_state :],
        )
        outputs_of_left = rg.squeeze(
            self.sys_left.get_observation(
                time=time,
                state=state_for_left,
                inputs=inputs_for_left,
                _native_dim=_native_dim,
            )
        )

        inputs_for_right = rg.zeros(
            self.sys_right.dim_inputs,
            prototype=(state, inputs),
        )
        if len(self.occupied_idx) > 0:
            inputs_for_right[self.occupied_idx] = outputs_of_left[self.rout_idx]
        inputs_for_right[self.free_right_input_indices] = rg.reshape(
            inputs[self.sys_left.dim_inputs :],
            rg.shape(inputs_for_right[self.free_right_input_indices]),
        )
        outputs_of_right = self.sys_right.get_observation(
            time=time,
            state=state_for_right,
            inputs=inputs_for_right,
            _native_dim=_native_dim,
        )
        if self.output_mode == "right":
            output = outputs_of_right
        elif self.output_mode == "state":
            output = state
        elif self.output_mode == "both":
            output = rg.concatenate((state_for_left, state_for_right))
        else:
            raise NotImplementedError

        ## TODO: implement 'preserve' mode

        return output

    def receive_action(self, action: np.array) -> None:
        self.inputs = action

    def update_system_parameters(self, inputs: Dict[str, float]) -> None:
        """Updates the parameters of both component subsystems based on a dictionary of new parameter values.

        This method propagates any system parameter updates to both the left and right subsystems that form the ComposedSystem.
        It is a convenient way to adjust or tune parameters affecting the dynamics and behavior of the overall system.

        Args:
            inputs: A dictionary where keys represent parameter names to update, and values represent the new parameter values.
            It is expected that both subsystems share a common parameter naming convention if the same parameters are to be updated across both systems.

        Note:
            Parameters that do not exist in either or both of the subsystems are ignored, and no exception is raised for such cases.
        """
        assert isinstance(inputs, dict)
        self.sys_left.update_system_parameters(inputs)
        self.sys_right.update_system_parameters(inputs)

    def permute_state(self, permutation: Union[List[int], np.array]) -> Self:
        """Permute an order at which the system outputs are returned.

        Args:
            permutation (Union[list, np.array]): Permutation represented
                as an array of indices

        Returns:
            Self: link to self
        """
        self.forward_permutation = rg.array(permutation).astype(int)
        self.inverse_permutation = self.get_inverse_permutation(permutation)
        return self

    def get_inverse_permutation(
        self, permutation: Union[List[int], np.array]
    ) -> np.array:
        """Calculate the inverse of a given permutation array.

        This method is particularly useful when a specific output ordering of the composed system's state vector
        needs to be reversed to its original form, such as when unpacking state vectors for separate subsystem processing.

        Args:
            permutation (Union[List[int], np.array]): The permutation array for which the inverse is to be computed.

        Returns:
            A numpy array representing the inverse permutation.

        Note:
            The given permutation array must represent a valid permutation of its indices for the inverse to be calculated correctly.
        """
        self.current_permutation = permutation
        permutation = np.asanyarray(permutation)
        inverse_permutation = np.empty_like(permutation)
        inverse_permutation[permutation] = np.arange(permutation.size)
        return inverse_permutation

    def compose(
        self,
        sys_right: SystemInterface,
        io_mapping: Optional[List[int]] = None,
        output_mode: str = "state",
    ) -> ComposedSystem:
        """Create a new ComposedSystem by appending another system to the current one.

        This method allows for creating multi-layered system structures by successively combining systems.
        The right system specified will receive a connection from the current one as defined by the input/output mapping.

        Args:
            sys_right: The system to compose with the current one on the right.
            io_mapping: An optional list defining the state-to-input connections between the current system and the right system.
            output_mode: Determines how to create observations for the resulting composed system. Defaults to `"state"`.

        Returns:
            A new instance of ComposedSystem containing the current system as the 'left' system and the specified right system as the 'right' system,
                and connected according to the specified mapping and output mode.
        """
        return ComposedSystem(
            self, sys_right, io_mapping=io_mapping, output_mode=output_mode
        )

    def __matmul__(self, sys_right: SystemInterface) -> Self:
        """Special method to create a new ComposedSystem via the `@` operator.

        Using the `@` operator constructs a new ComposedSystem with the current instance acting as the left system and the parameter system becoming the right system in the composition.
        By default, the connections between systems are determined based on the indices of their states and inputs if an explicit mapping is not provided.

        Args:
            sys_right: The system to be connected to the current system.

        Returns:
            A new ComposedSystem instance representing the composition of the current and right systems.

        Example:
            ```python
            composed_system = system1 @ system2  # Compose system1 with system2.
            ```
        """
        return self.compose(sys_right)


# @animation.StateAnimation.attach
@callback.DefaultAnimation.attach
class System(SystemInterface):
    """Class representing a controllable system with predefined dynamics and observations.

    Attributes and methods are inherited from SystemInterface. This class serves as a base implementation of a system
    with additional functionality to receive control inputs and update system parameters.
    """

    def __init__(
        self,
        system_parameters_init: Dict[str, float] = None,
        state_init: Optional[np.ndarray] = None,
        inputs_init: Optional[np.ndarray] = None,
    ):
        """Initialize an instance of a system with optional overrides for system parameters, initial state, and inputs.

        Args:
            system_parameters_init: A dictionary containing system parameters
                that override the default parameters.

            state_init: An initial state to set for the system. This argument is
                currently not utilized in the system's logic and serves as a placeholder for
                potential future functionality where a system's state could be set at instantiation.

            inputs_init: Initial inputs to set for the system. Similar to
                state_init, this argument is not currently used in the implementation and is
                maintained for potential future enhancements where system inputs can be
                initialized upon creation.

        Note:
            The current implementation of the system class treats systems as static entities.
            The `state_init` and `inputs_init` parameters are part of the method signature
            for potential future use. As of now, they have no effect on the system's behavior
            and are not used in the code.
        """
        if system_parameters_init is None:
            system_parameters_init = {}
        assert self.name is not None
        assert self.system_type is not None
        assert self.dim_state is not None
        assert self.dim_inputs is not None
        assert self.dim_observation is not None
        assert isinstance(
            system_parameters_init, dict
        ), "system_parameters_init should be a dict"

        if system_parameters_init:
            self._parameters.update(system_parameters_init)

        self.system_parameters_init = self._parameters

        if state_init is None:
            self.state = rg.zeros(self.dim_state)
        else:
            self.state = state_init

        if inputs_init is None:
            self.inputs = rg.zeros(self.dim_inputs)
        else:
            self.inputs = inputs_init

    def _get_observation(
        self, time: Union[float, cs.MX], state: RgArray, inputs: RgArray
    ) -> RgArray:
        """Return state vector by default.

        In many cases the observation is simply the state vector, so it is quite reasonable to just
        set the observation to the state vector by default.

        Args:
            time: The current simulation time or step.
            state: The current state vector of the system.
            inputs: The current input vector given to the system.

        Returns:
            An observation vector representing the perceived state of the system at the current time.

        Note:
            Implementing classes should define what makes up an observation in the context of their specific system.
            The method should aim to represent realistically perceivable quantities
        """
        return state

    def receive_action(self, action: np.array) -> None:
        self.inputs = action

    def update_system_parameters(self, inputs: Dict[str, float]) -> Dict[str, float]:
        """Simply updates `_parameters` dict."""
        assert isinstance(inputs, dict)
        self._parameters.update(inputs)
        return self.parameters

    def compose(
        self,
        sys_right: SystemInterface,
        io_mapping: Optional[List[int]] = None,
        output_mode: str = "state",
    ) -> ComposedSystem:
        """Compose the current system with another system.

        Args:
            sys_right: The system to be composed with.
            io_mapping: A list specifying the input-output mapping between the current system and sys_right. Defaults to None.
            output_mode: The output mode of the composed system. Defaults to "state".

        Returns:
            The composed system.
        """
        return ComposedSystem(
            self, sys_right, io_mapping=io_mapping, output_mode=output_mode
        )

    def __matmul__(self, sys_right: SystemInterface) -> Self:
        """Special method to create a new ComposedSystem via the `@` operator.

        Using the `@` operator constructs a new ComposedSystem with the current instance acting as the left system and the parameter system becoming the right system in the composition.
        By default, the connections between systems are determined based on the indices of their states and inputs if an explicit mapping is not provided.

        Args:
            sys_right: The system to be connected to the current system.

        Returns:
            A new ComposedSystem instance representing the composition of the current and right systems.

        Example:
            composed_system = system1 @ system2  # Compose system1 with system2.
        """
        return self.compose(sys_right)

@callback.OmnirobotAnimation.attach
class KinematicPoint(System):
    """System representing a simple 2D kinematic point (see [here](../systems/kin_point.md) for details)."""

    _name = "kinematic-point"
    _system_type = "diff_eqn"
    _dim_state = 2
    _dim_inputs = 2
    _dim_observation = 2
    _observation_naming = _state_naming = ["x [m]", "y [m]"]
    _inputs_naming = ["v_x [m/s]", "v_y [m/s]"]
    _action_bounds = [[-10.0, 10.0], [-10.0, 10.0]]

    def _compute_state_dynamics(
        self, time: Union[float, cs.MX], state: RgArray, inputs: RgArray
    ) -> RgArray:
        """Compute [right-hand side](../systems/kin_point.md#system-dynamics) of kinematic point.

        Args:
            time: Current time.
            state: Current state.
            inputs: Current control inputs (i. e. action).

        Returns:
            Right-hand side of the kinematic point.
        """

        Dstate = rg.zeros(
            self.dim_state,
            prototype=(state, inputs),
        )

        for i in range(rg.shape(inputs)[0]):
            Dstate[i] = inputs[i]

        return Dstate

@callback.PendulumAnimation.attach
class InvertedPendulum(System):
    """System representing an [inverted pendulum](../systems/inv_pendulum.md), with state representing angle and angular velocity."""

    _name = "inverted-pendulum"
    _system_type = "diff_eqn"
    _dim_state = 2
    _dim_inputs = 1
    _dim_observation = 2
    _parameters = {"m": 1, "g": 9.8, "l": 1}
    _observation_naming = _state_naming = ["angle [rad]", "angular velocity [rad/s]"]
    _inputs_naming = ["momentum [kg*m/s]"]
    _action_bounds = [[-20.0, 20.0]]

    def _compute_state_dynamics(
        self, time: Union[float, cs.MX], state: RgArray, inputs: RgArray
    ) -> RgArray:
        """Compute [right-hand side](../systems/inv_pendulum.md#system-dynamics) of the inverted pendulum."""

        Dstate = rg.zeros(
            self.dim_state,
            prototype=(state, inputs),
        )

        m, g, l = (
            self.parameters["m"],
            self.parameters["g"],
            self.parameters["l"],
        )

        Dstate[0] = state[1]
        Dstate[1] = g / l * rg.sin(state[0]) + inputs[0] / (m * l**2)

        return Dstate

@callback.ThreeWheeledRobotAnimation.attach
class ThreeWheeledRobotKinematic(System):
    r"""Implements the [kinematic three-wheeled robot](../systems/3wrobot_kin.md) (a.k.a. Brockett integrator)."""

    _name = "three-wheeled-robot-ni"
    _system_type = "diff_eqn"
    _dim_state = 3
    _dim_inputs = 2
    _dim_observation = 3
    _observation_naming = _state_naming = ["x [m]", "y [m]", "angle [rad]"]
    _inputs_naming = ["velocity [m/s]", "angular velocity [rad/s]"]
    _action_bounds = [[-25.0, 25.0], [-5.0, 5.0]]

    def _compute_state_dynamics(
        self, time: Union[float, cs.MX], state: RgArray, inputs: RgArray
    ) -> RgArray:
        """Compute [right-hand side](../systems/3wrobot_kin.md#system-dynamics) of the dynamic system.

        Args:
            time: Current time.
            state: Current state.
            inputs: Current control inputs (i. e. action).

        Returns:
            Right-hand side of the non-holonomic robot.
        """
        Dstate = rg.zeros(self.dim_state, prototype=(state, inputs))

        Dstate[0] = inputs[0] * rg.cos(state[2])
        Dstate[1] = inputs[0] * rg.sin(state[2])
        Dstate[2] = inputs[1]

        return Dstate

@callback.ThreeWheeledRobotAnimation.attach
class ThreeWheeledRobotDynamic(System):
    r"""Implements [dynamic three-wheeled robot](../systems/3wrobot_dyn.md)."""

    _name = "three-wheeled-robot"
    _system_type = "diff_eqn"
    _dim_state = 5
    _dim_inputs = 2
    _dim_observation = 5
    _parameters = {"m": 10, "I": 1}
    _observation_naming = _state_naming = [
        "x [m]",
        "y [m]",
        "angle [rad]",
        "l_velocity [m/s]",
        "angular_velocity [rad/s]",
    ]
    _inputs_naming = ["Force [kg*m/s^2]", "Momentum [kg*m/s]"]
    _action_bounds = [[-50.0, 50.0], [-10.0, 10.0]]

    def _compute_state_dynamics(
        self, time: Union[float, cs.MX], state: RgArray, inputs: RgArray
    ) -> RgArray:
        """Compute [right-hand side](../systems/3wrobot_dyn.md#system-dynamics) of the dynamic system.

        Args:
            time: Current time.
            state: Current state.
            inputs: Current action.

        Returns:
            Right-hand side.
        """
        Dstate = rg.zeros(
            self.dim_state,
            prototype=(state, inputs),
        )

        m, I = self.parameters["m"], self.parameters["I"]

        Dstate[0] = state[3] * rg.cos(state[2])
        Dstate[1] = state[3] * rg.sin(state[2])
        Dstate[2] = state[4]
        Dstate[3] = 1 / m * inputs[0]
        Dstate[4] = 1 / I * inputs[1]

        return Dstate


class Integrator(System):
    """System yielding non-holonomic double integrator when composed with [kinematic three-wheeled robot][regelum.system.ThreeWheeledRobotKinematic].

    See [here](../tutorials/systems.md#example-3-combining-integrator-with-kinematic-robot) for details.
    """

    _name = "integral-parts"
    _system_type = "diff_eqn"
    _dim_state = 2
    _dim_inputs = 2
    _dim_observation = 2
    _parameters = {"m": 10, "I": 1}
    _action_bounds = [[-50.0, 50.0], [-10.0, 10.0]]

    def _compute_state_dynamics(
        self, time: Union[float, cs.MX], state: RgArray, inputs: RgArray
    ) -> RgArray:
        """Compute [right-hand side](../tutorials/systems.md#example-3-combining-integrator-with-kinematic-robot) of the dynamic system.

        Args:
            time: Current time.
            state: Current state.
            inputs: Current action.

        Returns:
            Right-hand side.
        """
        Dstate = rg.zeros(
            self.dim_state,
            prototype=(state, inputs),
        )

        m, I = self.parameters["m"], self.parameters["I"]

        Dstate[0] = 1 / m * inputs[0]
        Dstate[1] = 1 / I * inputs[1]

        return Dstate

@callback.CartpoleAnimation.attach
class CartPole(System):
    """[Cart pole system](../systems/cartpole.md) without friction."""

    _name = "cartpole"
    _system_type = "diff_eqn"
    _dim_state = 4
    _dim_inputs = 1
    _dim_observation = 5
    _parameters = {"m_c": 0.1, "m_p": 2.0, "g": 9.81, "l": 0.5}
    _observation_naming = _state_naming = [
        "angle [rad]",
        "x [m]",
        "angle_dot [rad/s]",
        "x_dot [m/s]",
    ]
    _inputs_naming = ["force [kg*m/s^2]"]
    _action_bounds = [[-50.0, 50.0]]

    def _compute_state_dynamics(
        self, time: Union[float, cs.MX], state: RgArray, inputs: RgArray
    ) -> RgArray:
        """Compute [right-hand side](../systems/cartpole.md#system-dynamics) of the dynamic system.

        Args:
            time: Current time.
            state: Current state.
            inputs: Current action.

        Returns:
            Right-hand side.
        """
        Dstate = rg.zeros(
            self.dim_state,
            prototype=(state, inputs),
        )

        m_c, m_p, g, l = (
            self.parameters["m_c"],
            self.parameters["m_p"],
            self.parameters["g"],
            self.parameters["l"],
        )
        theta = state[0]
        theta_dot = state[2]
        x_dot = state[3]
        Force = inputs[0]
        sin_theta = rg.sin(theta)
        cos_theta = rg.cos(theta)
        theta_acc = (
            (
                g * sin_theta
                - cos_theta * (Force + m_p * l * theta_dot**2 * sin_theta) / (m_c + m_p)
            )
            / l
            / (4 / 3 - m_p * (cos_theta**2) / (m_c + m_p))
        )

        Dstate[0] = theta_dot
        Dstate[1] = x_dot
        Dstate[2] = theta_acc
        Dstate[3] = (
            Force + m_p * l * (theta_dot**2 * sin_theta - theta_acc * cos_theta)
        ) / (m_c + m_p)

        return Dstate

    def _get_observation(
        self, time: Union[float, cs.MX], state: RgArray, inputs: RgArray
    ):
        theta = state[0]
        x = state[1]
        theta_dot = state[2]
        x_dot = state[3]

        # theta_observed = theta - rg.floor(theta / (2 * np.pi)) * 2 * np.pi
        # theta_observed = rg.if_else(
        #     theta_observed > np.pi,
        #     theta_observed - 2 * np.pi,
        #     theta - rg.floor(theta / (2 * np.pi)) * 2 * np.pi,
        # )
        # return rg.hstack([theta_observed, x, theta_dot, x_dot])
        # return rg.hstack([rg.sin(theta), 1 - rg.cos(theta), theta_dot, x_dot])
        # return rg.hstack([rg.sin(theta), 1 - rg.cos(theta), x, theta_dot, x_dot])
        return rg.hstack([rg.sin(theta), 1 - rg.cos(theta), x, theta_dot, x_dot])


class CartPolePG(CartPole):
    _name = "cartpole_pg"
    _system_type = "diff_eqn"
    _dim_state = 4
    _dim_inputs = 1
    _dim_observation = 5
    _parameters = {"m_c": 0.1, "m_p": 2.0, "g": 9.81, "l": 0.5}
    _observation_naming = _state_naming = [
        "angle [rad]",
        "x [m]",
        "angle_dot [rad/s]",
        "x_dot [m/s]",
    ]
    _inputs_naming = ["force [kg*m/s^2]"]
    _action_bounds = [[-50.0, 50.0]]
    _dim_observation = 4

    def _get_observation(
        self, time: Union[float, cs.MX], state: RgArray, inputs: RgArray
    ):
        theta = state[0]
        x = state[1]
        theta_dot = state[2]
        x_dot = state[3]
        # theta_observed = theta - rg.floor(theta / (2 * np.pi)) * 2 * np.pi
        # theta_observed = rg.if_else(
        #     theta_observed > np.pi,
        #     theta_observed - 2 * np.pi,
        #     theta - rg.floor(theta / (2 * np.pi)) * 2 * np.pi,
        # )
        # return rg.hstack([theta_observed, x, theta_dot, x_dot])
        return rg.hstack([rg.sin(theta), 1 - rg.cos(theta), theta_dot, x_dot])
        # return rg.hstack([rg.sin(theta), 1 - rg.cos(theta), x, theta_dot, x_dot])


@callback.TwoTankAnimation.attach
class TwoTank(System):
    """This module simulates a [Two-Tank System](../systems/2tank.md).

    Notes:
        Please be aware that, despite the presence of a link to the [tutorial](../systems/2tank.md),
        this current system implementation does not replicate the tutorial's version.
        The primary distinction lies in the handling of observations: in this variant,
        observations are equivalent to the system's state variables.

        For an alternative version where the observation is differentiated from the state by a
        reference value (specifically, $(0.4, 0.4)$ subtracted from the state), please consult
        [`TwoTankReferenced`][regelum.system.TwoTankReferenced].
        It is this [`TwoTankReferenced`](regelum.system.TwoTankReferenced) model
        that faithfully reproduces the system outlined in the [tutorial](../systems/2tank.md).
    """

    _name = "two-tank"
    _system_type = "diff_eqn"
    _dim_state = 2
    _dim_inputs = 1
    _dim_observation = 2
    _parameters = {"tau1": 18.4, "tau2": 24.4, "K1": 1.3, "K2": 1.0, "K3": 0.2}
    _observation_naming = _state_naming = [
        "Intake Tank Level [m]",
        "Sink Tank Level [m]",
    ]
    _inputs_naming = ["Pressure [Pa]"]
    _action_bounds = [[0.0, 1.0]]

    def _compute_state_dynamics(
        self, time: Union[float, cs.MX], state: RgArray, inputs: RgArray
    ) -> RgArray:
        """Compute [right-hand side](../systems/2tank.md#system-dynamics) of the dynamic system.

        Notes:
            Please be aware that, despite the presence of a link to the [tutorial](../systems/2tank.md),
            this current system implementation does not replicate the tutorial's version.
            The primary distinction lies in the handling of observations: in this variant,
            observations are equivalent to the system's state variables.

            For an alternative version where the observation is differentiated from the state by a
            reference value (specifically, $(0.4, 0.4)$ subtracted from the state), please consult
            [`TwoTankReferenced`][regelum.system.TwoTankReferenced].
            It is this [`TwoTankReferenced`](regelum.system.TwoTankReferenced) model
            that faithfully reproduces the system outlined in the [tutorial](../systems/2tank.md).

        Args:
            time: Current time.
            state: Current state.
            inputs: Current action.

        Returns:
            Right-hand side.
        """
        tau1, tau2, K1, K2, K3 = (
            self.parameters["tau1"],
            self.parameters["tau2"],
            self.parameters["K1"],
            self.parameters["K2"],
            self.parameters["K3"],
        )

        Dstate = rg.zeros(
            self.dim_state,
            prototype=(state, inputs),
        )
        Dstate[0] = 1 / (tau1) * (-state[0] + K1 * inputs[0])
        Dstate[1] = 1 / (tau2) * (-state[1] + K2 * state[0] + K3 * state[1] ** 2)

        return Dstate

@callback.LunarLanderAnimation.attach
class LunarLander(System):
    """[Lunar lander system](../systems/lunar_lander.md).

    The basis of this system is taken from the [following
    paper](https://web.aeromech.usyd.edu.au/AMME3500/Course_documents/material/tutorials/Assignment%204%20Lunar%20Lander%20Solution.pdf).

    The LunarLander class simulates a lunar lander during its descent to the moon's surface. It models both the
    vertical and lateral dynamics as well as the lander's rotation.

    Notes:
        Please be aware that, despite the presence of a link to the [tutorial](../systems/lunar_lander.md),
        this current system implementation does not replicate the tutorial's version.
        The primary distinction lies in the handling of observations: in this variant,
        observations are equivalent to the system's state variables.

        For an alternative version where the observation is differentiated from the state by a
        reference value (specifically, $(0.4, 0.4)$ subtracted from the state), please consult
        [`LunarLanderReferenced`][regelum.system.LunarLanderReferenced].
        It is this [`LunarLanderReferenced`](regelum.system.LunarLanderReferenced) model
        that faithfully reproduces the system outlined in the [tutorial](../systems/lunar_lander.md).
    """

    _name = "lander"
    _system_type = "diff_eqn"
    _dim_state = 6
    _dim_inputs = 2
    _dim_observation = 6
    _parameters = {"m": 10, "J": 3.0, "g": 1.625, "a": 1.0, "r": 1.0, "sigma": 1.0}
    _observation_naming = _state_naming = [
        "x [m]",
        "y [m]",
        "theta [rad]",
        "x_dot [m/s]",
        "y_dot [m/s]",
        "theta_dot [rad/s]",
    ]
    _inputs_naming = ["vertical force [kg*m/s^2]", "side force [kg*m/s^2]"]
    _action_bounds = [[-100.0, 100.0], [-50.0, 50.0]]

    def __init__(self, *args, **kwargs):
        """Initialize an instance of LunarLander by specifying relevant physical parameters."""
        super().__init__(*args, **kwargs)
        self.alpha = np.arctan(self.parameters["a"] / self.parameters["r"])
        self.l = np.sqrt(self.parameters["a"] ** 2 + self.parameters["r"] ** 2)
        self.is_landed = False

    def _compute_state_dynamics(
        self, time: Union[float, cs.MX], state: RgArray, inputs: RgArray
    ) -> RgArray:
        """Compute the state dynamics of the lunar lander based on its physical model.

        This method calculates the derivative of the lander's state variables, considering the applied forces from
        thrusters and the gravitational forces. It checks for landing conditions to alter the dynamics accordingly,
        such as setting velocities to zero upon touching down.

        Notes:
            Please be aware that, despite the presence of a link to the [tutorial](../systems/lunar_lander.md),
            this current system implementation does not replicate the tutorial's version.
            The primary distinction lies in the handling of observations: in this variant,
            observations are equivalent to the system's state variables.

            For an alternative version where the observation is differentiated from the state by a
            reference value (specifically, $(0.4, 0.4)$ subtracted from the state), please consult
            [`LunarLanderReferenced`][regelum.system.LunarLanderReferenced].
            It is this [`LunarLanderReferenced`](regelum.system.LunarLanderReferenced) model
            that faithfully reproduces the system outlined in the [tutorial](../systems/lunar_lander.md).

        Args:
            time: The current simulation time.
            state: The current state of the lunar lander.
            inputs: The current inputs to the lander.

        Returns:
            The state vector's derivative with respect to time.
        """
        Dstate_before_landing = rg.zeros(
            self.dim_state,
            prototype=(state, inputs),
        )

        m, J, g = (
            self.parameters["m"],
            self.parameters["J"],
            self.parameters["g"],
        )

        theta = state[2]
        x_dot = state[3]
        y_dot = state[4]
        theta_dot = state[5]

        left_support, right_support = self._compute_supports_geometry(state[:2], theta)

        self.is_landed = (
            rg.if_else(left_support[1] <= 0, 1, 0)
            + rg.if_else(right_support[1] <= 0, 1, 0)
        ) > 0

        F_l = inputs[0] * (1 - self.is_landed)
        F_t = inputs[1] * (1 - self.is_landed)

        self.is_landed_left = rg.if_else(left_support[1] <= 0, 1, 0)
        self.is_landed_right = rg.if_else(right_support[1] <= 0, 1, 0)
        self.is_landed_vertex = rg.if_else(state[1] <= 0, 1, 0)
        self.is_freezed = (
            self.is_landed_left * self.is_landed_right + self.is_landed_vertex
        ) > 0
        self.is_landed = (
            self.is_landed_left + self.is_landed_right + self.is_landed_vertex
        ) > 0

        Dstate_before_landing[0] = x_dot
        Dstate_before_landing[1] = y_dot
        Dstate_before_landing[2] = theta_dot
        Dstate_before_landing[3] = 1 / m * (F_l * rg.cos(theta) - F_t * rg.sin(theta))
        Dstate_before_landing[4] = (
            1 / m * (F_l * rg.sin(theta) + F_t * rg.cos(theta)) - g
        )
        Dstate_before_landing[5] = (4 * F_l) / J

        Dstate_landed_right = self._compute_pendulum_dynamics(
            angle=-theta - self.alpha,
            angle_dot=theta_dot,
            prototype=(state, inputs),
        )

        Dstate_landed_left = self._compute_pendulum_dynamics(
            angle=self.alpha - theta,
            angle_dot=theta_dot,
            prototype=(state, inputs),
        )

        # Check if any of the two lander's supports touched the ground. If yes, freeze the state.

        Dstate = (1 - self.is_freezed) * (
            (1 - self.is_landed) * Dstate_before_landing
            + self.is_landed
            * (
                self.is_landed_right * Dstate_landed_right
                + self.is_landed_left * Dstate_landed_left
            )
        )

        return Dstate

    def _compute_pendulum_dynamics(
        self,
        angle: Union[RgArray, float],
        angle_dot: Union[RgArray, float],
        prototype: RgArray,
    ) -> RgArray:
        """Compute the [dynamics](../systems/lunar_lander.md#system-dynamics) of the lander when it is in contact with the ground, constrained like a pendulum.

        This method is used internally to modify the dynamics of the lander when one of its extremities touches down,
        causing it to behave like a pendulum.

        Notes:
            Please be aware that, despite the presence of a link to the [tutorial](../systems/lunar_lander.md),
            this current system implementation does not replicate the tutorial's version.
            The primary distinction lies in the handling of observations: in this variant,
            observations are equivalent to the system's state variables.

            For an alternative version where the observation is differentiated from the state by a
            reference value (specifically, $(0.4, 0.4)$ subtracted from the state), please consult
            [`LunarLanderReferenced`][regelum.system.LunarLanderReferenced].
            It is this [`LunarLanderReferenced`](regelum.system.LunarLanderReferenced) model
            that faithfully reproduces the system outlined in the [tutorial](../systems/lunar_lander.md).

        Args:
            angle: The angle of the pendulum.
            angle_dot: The derivative of the pendulum angle.
            prototype: A prototype array indicating the desired shape and type for the returned state derivative.

        Returns:
            The computed state vector's derivative with pendulum-like constraints applied.
        """
        Dstate = rg.zeros(
            self.dim_state,
            prototype=prototype,
        )
        g = self.parameters["g"]

        x = self.l * rg.sin(angle)
        y = self.l * rg.cos(angle)
        angular_acceleration = g / self.l**2 * x
        Dstate[5] = angular_acceleration
        Dstate[0] = angle_dot * y
        Dstate[1] = -angle_dot * x
        Dstate[2] = -angle_dot
        Dstate[3] = y * angular_acceleration - angle_dot**2 * x
        Dstate[4] = -x * angular_acceleration - angle_dot**2 * y

        return Dstate

    def _compute_supports_geometry(
        self, xi: Union[RgArray, float], theta: Union[RgArray, float]
    ) -> Tuple[Union[RgArray, float], Union[RgArray, float]]:
        """Calculate the positions of the left and right supports of the lander.

        Args:
            xi: The position vector of the lander's center of gravity.
            theta: The current orientation angle of the lander.

        Returns:
            The positions of the left and right supports.
        """
        rotation_operator = rg.zeros((2, 2), prototype=xi)
        xi_2 = rg.zeros((2, 1), prototype=xi)
        xi_3 = rg.zeros((2, 1), prototype=xi)

        rotation_operator[0, 0] = rg.cos(theta)
        rotation_operator[0, 1] = -rg.sin(theta)
        rotation_operator[1, 0] = rg.sin(theta)
        rotation_operator[1, 1] = rg.cos(theta)

        a, r = self.parameters["a"], self.parameters["r"]
        xi_2[0, 0] = xi[0, 0] - a
        xi_2[1, 0] = xi[1, 0] - r
        xi_3[0, 0] = xi[0, 0] + a
        xi_3[1, 0] = xi[1, 0] - r

        xi_2_d = xi_2 - xi
        xi_3_d = xi_3 - xi

        xi_2_d_rot = rotation_operator @ xi_2_d
        xi_3_d_rot = rotation_operator @ xi_3_d
        xi_2_new = xi + xi_2_d_rot
        xi_3_new = xi + xi_3_d_rot
        return xi_2_new, xi_3_new


@callback.detach
class ConstantReference(System):
    """Subtracts reference from system."""

    name = "constant_reference"
    _system_type = "diff_eqn"
    _dim_state = 0
    _dim_inputs = 2
    _dim_observation = 2
    _parameters = {"reference": np.array([[0.4], [0.4]])}

    def __init__(self, reference: Optional[Union[List[float], np.array]] = None):
        """Instantiate ConstantReference.

        Args:
            reference (Optional[Union[List[float], np.array]], optional):
                reference to be substracted from inputs, defaults to None
        """
        if reference is None:
            super().__init__(
                system_parameters_init=None, state_init=None, inputs_init=None
            )
        else:
            super().__init__(
                system_parameters_init={
                    "reference": np.array(reference).reshape(-1, 1)
                },
                state_init=None,
                inputs_init=None,
            )
            self._dim_inputs = self._dim_observation = (
                np.array(reference).reshape(-1).shape[0]
            )

    def _get_observation(
        self, time: Union[float, cs.MX], state: RgArray, inputs: RgArray
    ) -> RgArray:
        return inputs - rg.array(
            self.parameters["reference"], prototype=inputs, _force_numeric=True
        )

    def _compute_state_dynamics(
        self, time: Union[float, cs.MX], state: RgArray, inputs: RgArray
    ) -> RgArray:
        return inputs


class SystemWithConstantReference(ComposedSystem):
    """Composed system created by combining a system with a constant reference subtraction."""

    def __init__(self, system: System, state_reference: Union[List[float], np.array]):
        """Instantiate System with ConstantReference.

        The result ComposedSystem's method get_observation subtracts from state reference value state_reference.

        Args:
            system (System): system
            state_reference (Union[List[float], np.array]): reference to
                be subtracted from state.
        """
        constant_reference = ConstantReference(state_reference)

        assert (
            system.dim_state == constant_reference.dim_inputs
        ), "state_reference should have the same length as dimension of state of system"
        np.array(constant_reference).reshape(-1)
        super().__init__(
            sys_left=system,
            sys_right=constant_reference,
            io_mapping=None,
            output_mode="right",
            inputs_naming=system.inputs_naming,
            state_naming=system.state_naming,
            observation_naming=[s + "-ref" for i, s in enumerate(system.state_naming)],
            action_bounds=system.action_bounds,
        )


class LunarLanderReferenced(SystemWithConstantReference):
    """Lunar lander referenced system.

    This class represents a tailored version of the [`LunarLander`](regelum.system.LunarLander)
    with specific enhancements to match the instructional framework provided by the [tutorial](../systems/lunar_lander.md).

    Notes:
        The [`LunarLanderReferenced`][regelum.system.LunarLanderReferenced] system varies from the original
        [`LunarLander`][regelum.system.LunarLander] in one significant aspect: the computation of the observation.
        In [`LunarLander`][regelum.system.LunarLander], the observation directly equals the current state.
        However, in [`LunarLanderReferenced`][regelum.system.LunarLanderReferenced],
        the observation is calculated by subtracting a fixed reference value of $(0, 1, 0, 0, 0, 0)$ from the current state,
        resulting in an adjusted observation value. This modification
        is designed for scenarios where the observation needs to be anchored to a particular reference point.

        Despite this alteration in the observation representation, the core dynamics of the state remain unaltered.
        They follow the same principles as elucidated in the [system dynamics section of the tutorial](../systems/lunar_lander.md#system-dynamics)
        and are consistent with the [`LunarLander`][regelum.system.LunarLander] class.
        This ensures that while the observation method has been modified, the fundamental
        behavior of the system **state** dynamics stays true to the original model.
    """

    def __init__(self):
        """Instantiate TwoTankReferenced."""
        super().__init__(
            system=LunarLander(), state_reference=[0.0, 1.0, 0.0, 0.0, 0.0, 0.0]
        )


class TwoTankReferenced(SystemWithConstantReference):
    """Two tank referenced system.

    This class represents a tailored version of the [`TwoTank`][regelum.system.TwoTank]
    with specific enhancements to match the instructional framework provided by the [tutorial](../systems/2tank.md).

    Notes:
        The [`TwoTankReferenced`][regelum.system.TwoTankReferenced] system varies from the original
        [`TwoTank`][regelum.system.TwoTank] in one significant aspect: the computation of the observation.
        In [`TwoTank`][regelum.system.TwoTank], the observation directly equals the current state.
        However, in [`TwoTankReferenced`][regelum.system.TwoTankReferenced],
        the observation is calculated by subtracting a fixed reference value of $(0.4, 0.4)$ from the current state,
        resulting in an adjusted observation value. This modification
        is designed for scenarios where the observation needs to be anchored to a particular reference point.

        Despite this alteration in the observation representation, the core dynamics of the state remain unaltered.
        They follow the same principles as elucidated in the [system dynamics section of the tutorial](../systems/2tank.md#system-dynamics)
        and are consistent with the [`TwoTank`][regelum.system.TwoTank] class.
        This ensures that while the observation method has been modified, the fundamental
        behavior of the system **state** dynamics stays true to the original model.

    """

    def __init__(self):
        """Instantiate TwoTankReferenced."""
        super().__init__(system=TwoTank(), state_reference=[0.4, 0.4])



"""Type Specifications for the Regelum Framework.

This module contains type aliases for array types that are used throughout the Regelum framework.
These type specifications are utilized to ensure that the framework can handle different kinds
of array-like data structures consistently across its various components, providing flexibility
and compatibility with major numerical computing libraries.

The type specifications are designed to support the integration of Numpy arrays, PyTorch tensors,
and CasADi matrix types, making it easier to develop reinforcement learning and optimal control
algorithms that can leverage the strengths of each of these libraries.
"""

import torch
import numpy as np
import casadi as cs
from typing import Union, Type, OrderedDict


RgArrayType = Union[
    Type[np.array],
    Type[torch.Tensor],
    Type[cs.DM],
    Type[cs.MX],
]
"""
Type alias for the classes of array types supported by the Regelum framework.

This type alias includes the classes, not instances, of the array types that can be used within
the Regelum framework for RL and optimal control computations. 

Attributes:
    np.array: Represents the class of Numpy array.
    torch.Tensor: Represents the class of PyTorch tensor.
    cs.DM: Represents the class of CasADi's DM matrix type.
    cs.MX: Represents the class of CasADi's MX matrix type.
"""

RgArray = Union[
    np.array,
    torch.Tensor,
    cs.DM,
    cs.MX,
]
"""
Type alias for the instances of array types supported by the Regelum framework.

This type alias includes the instances of the array types that can be used within
the Regelum framework for RL and optimal control computations.

Attributes:
    np.array: An instance of a Numpy array.
    torch.Tensor: An instance of a PyTorch tensor.
    cs.DM: An instance of CasADi's DM matrix.
    cs.MX: An instance of CasADi's MX matrix.
"""

Weights = Union[RgArray, OrderedDict[str, torch.Tensor]]
"""Type alias for the instances of model's weights. For ModelNN weights are `OrderedDict[str, torch.Tensor]`. For Model weights are `RgArray`."""
